import os

from django.urls import reverse
from PIL import Image

from ...acl.models import Role
from ...acl.test import patch_user_acl
from ...conf import settings
from ...users.test import AuthenticatedUserTestCase
from ..models import Attachment, AttachmentType

TESTFILES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "testfiles")
TEST_DOCUMENT_PATH = os.path.join(TESTFILES_DIR, "document.pdf")
TEST_LARGEPNG_PATH = os.path.join(TESTFILES_DIR, "large.png")
TEST_SMALLJPG_PATH = os.path.join(TESTFILES_DIR, "small.jpg")
TEST_ANIMATEDGIF_PATH = os.path.join(TESTFILES_DIR, "animated.gif")
TEST_CORRUPTEDIMG_PATH = os.path.join(TESTFILES_DIR, "corrupted.gif")


class AttachmentsApiTestCase(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        AttachmentType.objects.all().delete()

        self.api_link = reverse("misago:api:attachment-list")

    def test_anonymous(self):
        """user has to be authenticated to be able to upload files"""
        self.logout_user()

        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)

    @patch_user_acl({"max_attachment_size": 0})
    def test_no_permission(self):
        """user needs permission to upload files"""
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You don't have permission to upload new files."},
        )

    def test_no_file_uploaded(self):
        """no file uploaded scenario is handled"""
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "No file has been uploaded."})

    def test_invalid_extension(self):
        """uploaded file's extension is rejected as invalid"""
        AttachmentType.objects.create(
            name="Test extension", extensions="jpg,jpeg", mimetypes=None
        )

        with open(TEST_DOCUMENT_PATH, "rb") as upload:
            response = self.client.post(self.api_link, data={"upload": upload})
            self.assertEqual(response.status_code, 400)
            self.assertEqual(
                response.json(), {"detail": "You can't upload files of this type."}
            )

    def test_invalid_mime(self):
        """uploaded file's mimetype is rejected as invalid"""
        AttachmentType.objects.create(
            name="Test extension", extensions="png", mimetypes="loremipsum"
        )

        with open(TEST_DOCUMENT_PATH, "rb") as upload:
            response = self.client.post(self.api_link, data={"upload": upload})
            self.assertEqual(response.status_code, 400)
            self.assertEqual(
                response.json(), {"detail": "You can't upload files of this type."}
            )

    def test_no_perm_to_type(self):
        """user needs permission to upload files of this type"""
        attachment_type = AttachmentType.objects.create(
            name="Test extension", extensions="png", mimetypes="application/pdf"
        )

        user_roles = (r.pk for r in self.user.get_roles())
        attachment_type.limit_uploads_to.set(Role.objects.exclude(id__in=user_roles))

        with open(TEST_DOCUMENT_PATH, "rb") as upload:
            response = self.client.post(self.api_link, data={"upload": upload})
            self.assertEqual(response.status_code, 400)
            self.assertEqual(
                response.json(), {"detail": "You can't upload files of this type."}
            )

    def test_type_is_locked(self):
        """new uploads for this filetype are locked"""
        AttachmentType.objects.create(
            name="Test extension",
            extensions="png",
            mimetypes="application/pdf",
            status=AttachmentType.LOCKED,
        )

        with open(TEST_DOCUMENT_PATH, "rb") as upload:
            response = self.client.post(self.api_link, data={"upload": upload})
            self.assertEqual(response.status_code, 400)
            self.assertEqual(
                response.json(), {"detail": "You can't upload files of this type."}
            )

    def test_type_is_disabled(self):
        """new uploads for this filetype are disabled"""
        AttachmentType.objects.create(
            name="Test extension",
            extensions="png",
            mimetypes="application/pdf",
            status=AttachmentType.DISABLED,
        )

        with open(TEST_DOCUMENT_PATH, "rb") as upload:
            response = self.client.post(self.api_link, data={"upload": upload})
            self.assertEqual(response.status_code, 400)
            self.assertEqual(
                response.json(), {"detail": "You can't upload files of this type."}
            )

    @patch_user_acl({"max_attachment_size": 100})
    def test_upload_too_big_for_user(self):
        """too big uploads are rejected"""
        AttachmentType.objects.create(
            name="Test extension", extensions="png", mimetypes="image/png"
        )

        with open(TEST_LARGEPNG_PATH, "rb") as upload:
            response = self.client.post(self.api_link, data={"upload": upload})
            self.assertEqual(response.status_code, 400)
            self.assertEqual(
                response.json(),
                {
                    "detail": (
                        "You can't upload files larger than 100.0\xa0KB "
                        "(your file has 253.9\xa0KB)."
                    )
                },
            )

    def test_corrupted_image_upload(self):
        """corrupted image upload is handled"""
        AttachmentType.objects.create(name="Test extension", extensions="gif")

        with open(TEST_CORRUPTEDIMG_PATH, "rb") as upload:
            response = self.client.post(self.api_link, data={"upload": upload})
            self.assertEqual(response.status_code, 400)
            self.assertEqual(
                response.json(), {"detail": "Uploaded image is unsupported or invalid."}
            )

    def test_document_upload(self):
        """successful upload creates orphan attachment"""
        AttachmentType.objects.create(
            name="Test extension", extensions="pdf", mimetypes="application/pdf"
        )

        with open(TEST_DOCUMENT_PATH, "rb") as upload:
            response = self.client.post(self.api_link, data={"upload": upload})
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        attachment = Attachment.objects.get(id=response_json["id"])

        self.assertEqual(attachment.filename, "document.pdf")
        self.assertTrue(attachment.is_file)
        self.assertFalse(attachment.is_image)

        self.assertIsNotNone(attachment.file)
        self.assertTrue(not attachment.image)
        self.assertTrue(not attachment.thumbnail)

        self.assertTrue(str(attachment.file).endswith("document.pdf"))

        self.assertIsNone(response_json["post"])
        self.assertEqual(response_json["uploader_name"], self.user.username)
        self.assertEqual(response_json["url"]["index"], attachment.get_absolute_url())
        self.assertIsNone(response_json["url"]["thumb"])
        self.assertEqual(response_json["url"]["uploader"], self.user.get_absolute_url())

        self.assertEqual(self.user.audittrail_set.count(), 1)

        # files associated with attachment are deleted on its deletion
        file_path = attachment.file.path
        self.assertTrue(os.path.exists(file_path))
        attachment.delete()
        self.assertFalse(os.path.exists(file_path))

    def test_small_image_upload(self):
        """successful small image upload creates orphan attachment without thumbnail"""
        AttachmentType.objects.create(
            name="Test extension", extensions="jpeg,jpg", mimetypes="image/jpeg"
        )

        with open(TEST_SMALLJPG_PATH, "rb") as upload:
            response = self.client.post(self.api_link, data={"upload": upload})
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        attachment = Attachment.objects.get(id=response_json["id"])

        self.assertEqual(attachment.filename, "small.jpg")
        self.assertFalse(attachment.is_file)
        self.assertTrue(attachment.is_image)

        self.assertTrue(not attachment.file)
        self.assertIsNotNone(attachment.image)
        self.assertTrue(not attachment.thumbnail)

        self.assertTrue(str(attachment.image).endswith("small.jpg"))

        self.assertIsNone(response_json["post"])
        self.assertEqual(response_json["uploader_name"], self.user.username)
        self.assertEqual(response_json["url"]["index"], attachment.get_absolute_url())
        self.assertIsNone(response_json["url"]["thumb"])
        self.assertEqual(response_json["url"]["uploader"], self.user.get_absolute_url())

        self.assertEqual(self.user.audittrail_set.count(), 1)

    @patch_user_acl({"max_attachment_size": 10 * 1024})
    def test_large_image_upload(self):
        """successful large image upload creates orphan attachment with thumbnail"""
        AttachmentType.objects.create(
            name="Test extension", extensions="png", mimetypes="image/png"
        )

        with open(TEST_LARGEPNG_PATH, "rb") as upload:
            response = self.client.post(self.api_link, data={"upload": upload})
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        attachment = Attachment.objects.get(id=response_json["id"])

        self.assertEqual(attachment.filename, "large.png")
        self.assertFalse(attachment.is_file)
        self.assertTrue(attachment.is_image)

        self.assertTrue(not attachment.file)
        self.assertIsNotNone(attachment.image)
        self.assertIsNotNone(attachment.thumbnail)

        self.assertTrue(str(attachment.image).endswith("large.png"))
        self.assertTrue(str(attachment.thumbnail).endswith("large.png"))

        self.assertIsNone(response_json["post"])
        self.assertEqual(response_json["uploader_name"], self.user.username)
        self.assertEqual(response_json["url"]["index"], attachment.get_absolute_url())
        self.assertEqual(response_json["url"]["thumb"], attachment.get_thumbnail_url())
        self.assertEqual(response_json["url"]["uploader"], self.user.get_absolute_url())

        self.assertEqual(self.user.audittrail_set.count(), 1)

        # thumbnail was scaled down
        thumbnail = Image.open(attachment.thumbnail.path)
        self.assertEqual(
            thumbnail.size[0], settings.MISAGO_ATTACHMENT_IMAGE_SIZE_LIMIT[0]
        )
        self.assertLess(
            thumbnail.size[1], settings.MISAGO_ATTACHMENT_IMAGE_SIZE_LIMIT[1]
        )

        # files associated with attachment are deleted on its deletion
        image_path = attachment.image.path
        thumbnail_path = attachment.thumbnail.path

        self.assertTrue(os.path.exists(image_path))
        self.assertTrue(os.path.exists(thumbnail_path))

        attachment.delete()

        self.assertFalse(os.path.exists(image_path))
        self.assertFalse(os.path.exists(thumbnail_path))

    def test_animated_image_upload(self):
        """successful gif upload creates orphan attachment with thumbnail"""
        AttachmentType.objects.create(
            name="Test extension", extensions="gif", mimetypes="image/gif"
        )

        with open(TEST_ANIMATEDGIF_PATH, "rb") as upload:
            response = self.client.post(self.api_link, data={"upload": upload})
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        attachment = Attachment.objects.get(id=response_json["id"])

        self.assertEqual(attachment.filename, "animated.gif")
        self.assertFalse(attachment.is_file)
        self.assertTrue(attachment.is_image)

        self.assertTrue(not attachment.file)
        self.assertIsNotNone(attachment.image)
        self.assertIsNotNone(attachment.thumbnail)

        self.assertTrue(str(attachment.image).endswith("animated.gif"))
        self.assertTrue(str(attachment.thumbnail).endswith("animated.gif"))

        self.assertIsNone(response_json["post"])
        self.assertEqual(response_json["uploader_name"], self.user.username)
        self.assertEqual(response_json["url"]["index"], attachment.get_absolute_url())
        self.assertEqual(response_json["url"]["thumb"], attachment.get_thumbnail_url())
        self.assertEqual(response_json["url"]["uploader"], self.user.get_absolute_url())

        self.assertEqual(self.user.audittrail_set.count(), 1)



from datetime import timedelta

from django.utils import timezone

from ..bans import get_user_ban
from ..models import BanCache, Online

ACTIVITY_CUTOFF = timedelta(minutes=2)


def make_users_status_aware(request, users, fetch_state=False):
    users_dict = {}
    for user in users:
        users_dict[user.pk] = user

    if fetch_state:
        # Fill ban cache on users
        for ban_cache in BanCache.objects.filter(user__in=users_dict.keys()):
            users_dict[ban_cache.user_id].ban_cache = ban_cache

        # Fill user online trackers
        for online_tracker in Online.objects.filter(user__in=users_dict.keys()):
            users_dict[online_tracker.user_id].online_tracker = online_tracker

    # Fill user states
    for user in users:
        user.status = get_user_status(request, user)


def get_user_status(request, user):
    user_status = {
        "is_banned": False,
        "is_hidden": user.is_hiding_presence,
        "is_online_hidden": False,
        "is_offline_hidden": False,
        "is_online": False,
        "is_offline": False,
        "banned_until": None,
        "last_click": user.last_login or user.joined_on,
    }

    user_ban = get_user_ban(user, request.cache_versions)
    if user_ban:
        user_status["is_banned"] = True
        user_status["banned_until"] = user_ban.expires_on

    try:
        online_tracker = user.online_tracker
        is_hidden = (
            user.is_hiding_presence and not request.user_acl["can_see_hidden_users"]
        )

        if online_tracker and not is_hidden:
            if online_tracker.last_click >= timezone.now() - ACTIVITY_CUTOFF:
                user_status["is_online"] = True
                user_status["last_click"] = online_tracker.last_click
    except Online.DoesNotExist:
        pass

    if user_status["is_hidden"]:
        if request.user_acl["can_see_hidden_users"]:
            user_status["is_hidden"] = False
            if user_status["is_online"]:
                user_status["is_online_hidden"] = True
                user_status["is_online"] = False
            else:
                user_status["is_offline_hidden"] = True
                user_status["is_offline"] = False
        else:
            user_status["is_hidden"] = True
    else:
        if user_status["is_online"]:
            user_status["is_online"] = True
        else:
            user_status["is_offline"] = True

    return user_status



from django.contrib.messages.api import get_messages
from django.contrib.messages.constants import ERROR, INFO, SUCCESS
from django.test import Client


def assert_contains(response, string, status_code=200):
    assert response.status_code == status_code
    fail_message = f'"{string}" not found in response.content'
    assert string in response.content.decode("utf-8"), fail_message


def assert_not_contains(response, string, status_code=200):
    assert response.status_code == status_code
    fail_message = f'"{string}" was unexpectedly found in response.content'
    assert string not in response.content.decode("utf-8"), fail_message


def assert_has_error_message(response):
    messages = get_messages(response.wsgi_request)
    levels = [i.level for i in messages]

    assert levels, "No messages were set during the request"
    assert ERROR in levels, "No error messages were set during the request"


def assert_has_info_message(response):
    messages = get_messages(response.wsgi_request)
    levels = [i.level for i in messages]

    assert levels, "No messages were set during the request"
    assert INFO in levels, "No info messages were set during the request"


def assert_has_success_message(response):
    messages = get_messages(response.wsgi_request)
    levels = [i.level for i in messages]

    assert levels, "No messages were set during the request"
    assert SUCCESS in levels, "No success messages were set during the request"


def assert_has_message(response, message, level=None):
    messages = get_messages(response.wsgi_request)
    found = False
    for msg in messages:
        if message in str(msg):
            if level and level != msg.level:
                error = (
                    'Message containing "%s" was set '
                    "but didn't have level %s (it had %s)"
                )
                raise AssertionError(error % (message, level, message.level))
            found = True

    if not found:
        raise AssertionError(
            'Message containing "%s" was not set during the request' % message
        )


class MisagoClient(Client):
    def post(self, *args, **kwargs):
        if "json" in kwargs:
            return super().post(
                *args,
                data=kwargs.pop("json"),
                content_type="application/json",
                **kwargs,
            )

        return super().post(*args, **kwargs)

    def put(self, *args, **kwargs):
        if "json" in kwargs:
            return super().put(
                *args,
                data=kwargs.pop("json"),
                content_type="application/json",
                **kwargs,
            )

        return super().put(*args, **kwargs)



import responses

from ..tasks import update_remote_css_size


@responses.activate
def test_task_uses_response_body_to_set_css_size(css_link):
    content = "html {}"
    content_bytes = content.encode()

    responses.add(
        responses.GET,
        css_link.url,
        headers={
            "Content-Type": "text/css;charset=utf-8",
            "Content-Length": str(len(content_bytes)),
        },
        body=content_bytes,
    )

    update_remote_css_size(css_link.pk)

    css_link.refresh_from_db()
    assert css_link.size == len(content)


@responses.activate
def test_task_doesnt_change_css_size_if_http_request_failed(css_link):
    responses.add(responses.GET, css_link.url, status=404)
    update_remote_css_size(css_link.pk)


def test_task_handles_css_without_url(css):
    update_remote_css_size(css.pk)


def test_task_handles_nonexisting_css(db):
    update_remote_css_size(1)



import json
import logging
import re

import requests
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.core.validators import EmailValidator
from django.utils.encoding import force_str
from django.utils.module_loading import import_string
from django.utils.translation import npgettext, pgettext_lazy
from requests.exceptions import RequestException

from .. import hooks
from ..conf import settings
from .bans import get_email_ban, get_username_ban

USERNAME_RE = re.compile(r"^[0-9a-z_]+$", re.IGNORECASE)
USERNAME_LATIN_ALPHA_RE = re.compile(r"[0-9a-z]", re.IGNORECASE)

logger = logging.getLogger("misago.users.validators")

User = get_user_model()


# E-mail validators
dj_validate_email = EmailValidator(
    message=pgettext_lazy("email validator", "Enter a valid e-mail address.")
)


def validate_email(value, exclude=None):
    """shortcut function that does complete validation of email"""
    dj_validate_email(value)
    validate_email_available(value, exclude)
    validate_email_banned(value)


def validate_email_available(value, exclude=None):
    try:
        user = User.objects.get_by_email(value)
        if not exclude or user.pk != exclude.pk:
            raise ValidationError(
                pgettext_lazy(
                    "user email validator", "This e-mail address is not available."
                )
            )
    except User.DoesNotExist:
        pass


def validate_email_banned(value):
    ban = get_email_ban(value, registration_only=True)

    if ban:
        if ban.user_message:
            raise ValidationError(ban.user_message)
        else:
            raise ValidationError(
                pgettext_lazy(
                    "user email validator", "This e-mail address is not allowed."
                )
            )


# Username validators
def validate_username(settings, value, exclude=None):
    """shortcut function that does complete validation of username"""
    validate_username_length(settings, value)
    validate_username_content(value)
    validate_username_available(value, exclude)
    validate_username_banned(value)


def validate_username_available(value, exclude=None):
    try:
        user = User.objects.get_by_username(value)
        if not exclude or user.pk != exclude.pk:
            raise ValidationError(
                pgettext_lazy("username validator", "This username is not available.")
            )
    except User.DoesNotExist:
        pass


def validate_username_banned(value):
    ban = get_username_ban(value, registration_only=True)

    if ban:
        if ban.user_message:
            raise ValidationError(ban.user_message)
        else:
            raise ValidationError(
                pgettext_lazy("username validator", "This username is not allowed.")
            )


def validate_username_content(value):
    if not USERNAME_RE.match(value):
        raise ValidationError(
            pgettext_lazy(
                "username validator",
                "Username can only contain Latin alphabet letters, digits, and an underscore sign.",
            )
        )

    if not USERNAME_LATIN_ALPHA_RE.search(value):
        raise ValidationError(
            pgettext_lazy(
                "username validator",
                "Username must contain Latin alphabet letters or digits.",
            )
        )


def validate_username_length(settings, value):
    if len(value) < settings.username_length_min:
        message = npgettext(
            "username length validator",
            "Username must be at least %(limit_value)s character long.",
            "Username must be at least %(limit_value)s characters long.",
            settings.username_length_min,
        )
        raise ValidationError(message % {"limit_value": settings.username_length_min})

    if len(value) > settings.username_length_max:
        message = npgettext(
            "username length validator",
            "Username cannot be longer than %(limit_value)s characters.",
            "Username cannot be longer than %(limit_value)s characters.",
            settings.username_length_max,
        )
        raise ValidationError(message % {"limit_value": settings.username_length_max})


# New account validators
SFS_API_URL = (
    "http://api.stopforumspam.org/api?email=%(email)s&ip=%(ip)s&f=json&confidence"
)


def validate_with_sfs(request, cleaned_data, add_error):
    if request.settings.enable_stop_forum_spam and cleaned_data.get("email"):
        try:
            _real_validate_with_sfs(
                request.user_ip,
                cleaned_data["email"],
                request.settings.stop_forum_spam_confidence,
            )
        except RequestException as error:
            logger.exception(error, exc_info=error)


def _real_validate_with_sfs(ip, email, confidence):
    r = requests.get(SFS_API_URL % {"email": email, "ip": ip}, timeout=5)
    r.raise_for_status()

    api_response = json.loads(force_str(r.content))
    ip_score = api_response.get("ip", {}).get("confidence", 0)
    email_score = api_response.get("email", {}).get("confidence", 0)

    api_score = max((ip_score, email_score))

    if api_score > confidence:
        raise ValidationError(
            pgettext_lazy(
                "stop forum spam validator",
                "Data entered was found in spammers database.",
            )
        )


def validate_gmail_email(request, cleaned_data, add_error):
    email = cleaned_data.get("email", "")
    if "@" not in email:
        return

    username, domain = email.lower().split("@")
    if domain == "gmail.com" and username.count(".") > 5:
        add_error(
            "email",
            ValidationError(
                pgettext_lazy("gmail email validator", "This e-mail is not allowed.")
            ),
        )


# Registration validation
validators_list = settings.MISAGO_NEW_REGISTRATIONS_VALIDATORS
REGISTRATION_VALIDATORS = list(map(import_string, validators_list))


def raise_validation_error(*_):
    raise ValidationError("")  # Raised when message content can be discarded


def validate_new_registration(request, cleaned_data, add_error=None, validators=None):
    validators = validators or REGISTRATION_VALIDATORS

    add_error = add_error or raise_validation_error
    for validator in validators:
        validator(request, cleaned_data, add_error)
    for validator in hooks.new_registrations_validators:
        validator(request, cleaned_data, add_error)



from django.contrib.auth import get_user_model
from django.urls import reverse
from rest_framework import serializers

from ...acl.useracl import serialize_user_acl
from .user import UserSerializer

User = get_user_model()

__all__ = ["AuthenticatedUserSerializer", "AnonymousUserSerializer"]


class AuthFlags:
    def get_is_authenticated(self, obj):
        return bool(obj.is_authenticated)

    def get_is_anonymous(self, obj):
        return bool(obj.is_anonymous)


class AuthenticatedUserSerializer(UserSerializer, AuthFlags):
    email = serializers.SerializerMethodField()
    is_authenticated = serializers.SerializerMethodField()
    is_anonymous = serializers.SerializerMethodField()

    class Meta:
        model = User
        fields = UserSerializer.Meta.fields + [
            "has_usable_password",
            "is_hiding_presence",
            "limits_private_thread_invites_to",
            "unread_private_threads",
            "watch_started_threads",
            "watch_replied_threads",
            "watch_new_private_threads_by_followed",
            "watch_new_private_threads_by_other_users",
            "notify_new_private_threads_by_followed",
            "notify_new_private_threads_by_other_users",
            "is_authenticated",
            "is_anonymous",
        ]

    def get_acl(self, obj):
        acl = self.context.get("acl")
        if acl:
            return serialize_user_acl(acl)
        return {}

    def get_email(self, obj):
        return obj.email

    def get_api(self, obj):
        return {
            "avatar": reverse("misago:api:user-avatar", kwargs={"pk": obj.pk}),
            "data_downloads": reverse(
                "misago:api:user-data-downloads", kwargs={"pk": obj.pk}
            ),
            "details": reverse("misago:api:user-details", kwargs={"pk": obj.pk}),
            "change_email": reverse(
                "misago:api:user-change-email", kwargs={"pk": obj.pk}
            ),
            "change_password": reverse(
                "misago:api:user-change-password", kwargs={"pk": obj.pk}
            ),
            "edit_details": reverse(
                "misago:api:user-edit-details", kwargs={"pk": obj.pk}
            ),
            "options": reverse("misago:api:user-forum-options", kwargs={"pk": obj.pk}),
            "request_data_download": reverse(
                "misago:api:user-request-data-download", kwargs={"pk": obj.pk}
            ),
            "username": reverse("misago:api:user-username", kwargs={"pk": obj.pk}),
            "delete": reverse(
                "misago:api:user-delete-own-account", kwargs={"pk": obj.pk}
            ),
        }

    def to_representation(self, instance) -> dict:
        data = super().to_representation(instance)
        data["unreadNotifications"] = instance.get_unread_notifications_for_display()
        return data


AuthenticatedUserSerializer = AuthenticatedUserSerializer.exclude_fields(
    "is_avatar_locked",
    "is_blocked",
    "is_followed",
    "is_signature_locked",
    "meta",
    "signature",
    "status",
)


class AnonymousUserSerializer(serializers.Serializer, AuthFlags):
    id = serializers.ReadOnlyField()
    acl = serializers.SerializerMethodField()
    is_authenticated = serializers.SerializerMethodField()
    is_anonymous = serializers.SerializerMethodField()

    def get_acl(self, obj):
        acl = self.context.get("acl")
        if acl:
            return serialize_user_acl(acl)
        return {}



from ...acl.test import patch_user_acl
from ..test import AuthenticatedUserTestCase


class UsernameChangesApiTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()
        self.link = "/api/username-changes/"

    @patch_user_acl({"can_see_users_name_history": False})
    def test_user_can_always_see_his_name_changes(self):
        """list returns own username changes"""
        self.user.set_username("NewUsername", self.user)
        response = self.client.get("%s?user=%s" % (self.link, self.user.pk))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, self.user.username)

    @patch_user_acl({"can_see_users_name_history": True})
    def test_list_handles_invalid_filter(self):
        """list raises 404 for invalid filter"""
        self.user.set_username("NewUsername", self.user)
        response = self.client.get("%s?user=abcd" % self.link)
        self.assertEqual(response.status_code, 404)

    @patch_user_acl({"can_see_users_name_history": True})
    def test_list_handles_nonexisting_user(self):
        """list raises 404 for invalid user id"""
        self.user.set_username("NewUsername", self.user)
        response = self.client.get("%s?user=142141" % self.link)
        self.assertEqual(response.status_code, 404)

    @patch_user_acl({"can_see_users_name_history": False})
    def test_list_handles_search(self):
        """list returns found username changes"""
        self.user.set_username("NewUsername", self.user)

        response = self.client.get("%s?user=%s&search=new" % (self.link, self.user.pk))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, self.user.username)

        response = self.client.get(
            "%s?user=%s&search=usernew" % (self.link, self.user.pk)
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["count"], 0)

    @patch_user_acl({"can_see_users_name_history": False})
    def test_list_denies_permission(self):
        """list denies permission for other user (or all) if no access"""
        response = self.client.get("%s?user=%s" % (self.link, self.user.pk + 1))
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You don't have permission to see other users name history."},
        )

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You don't have permission to see other users name history."},
        )



from importlib.util import find_spec

from django.urls import URLResolver, include, path


def discover_plugins_urlpatterns(plugins: list[str]) -> list[URLResolver]:
    urlpatterns: list[URLResolver] = []
    for plugin in plugins:
        if plugin_has_urls(plugin):
            urlpatterns.append(path("", include(f"{plugin}.urls")))

    return urlpatterns


def plugin_has_urls(plugin: str) -> bool:
    return bool(find_spec(f"{plugin}.urls"))



from django.core.exceptions import PermissionDenied, ValidationError
from django.template.defaultfilters import filesizeformat
from django.utils.translation import pgettext
from rest_framework import viewsets
from rest_framework.response import Response

from ...acl.objectacl import add_acl_to_obj
from ...users.audittrail import create_audit_trail
from ..models import Attachment, AttachmentType
from ..serializers import AttachmentSerializer

IMAGE_EXTENSIONS = ("jpg", "jpeg", "png", "gif", "webp")


class AttachmentViewSet(viewsets.ViewSet):
    def create(self, request):
        if not request.user_acl["max_attachment_size"]:
            raise PermissionDenied(
                pgettext(
                    "attachments api", "You don't have permission to upload new files."
                )
            )

        try:
            return self.create_attachment(request)
        except ValidationError as e:
            return Response({"detail": e.args[0]}, status=400)

    def create_attachment(self, request):
        upload = request.FILES.get("upload")
        if not upload:
            raise ValidationError(
                pgettext("attachments api", "No file has been uploaded.")
            )

        user_roles = set(r.pk for r in request.user.get_roles())
        filetype = validate_filetype(upload, user_roles)
        validate_filesize(upload, request.user_acl["max_attachment_size"])

        attachment = Attachment(
            secret=Attachment.generate_new_secret(),
            filetype=filetype,
            size=upload.size,
            uploader=request.user,
            uploader_name=request.user.username,
            uploader_slug=request.user.slug,
            filename=upload.name,
        )

        if is_upload_image(upload):
            try:
                attachment.set_image(upload)
            except IOError:
                raise ValidationError(
                    pgettext(
                        "attachments api", "Uploaded image is unsupported or invalid."
                    )
                )
        else:
            attachment.set_file(upload)

        attachment.save()
        add_acl_to_obj(request.user_acl, attachment)

        create_audit_trail(request, attachment)

        return Response(
            AttachmentSerializer(attachment, context={"user": request.user}).data
        )


def validate_filetype(upload, user_roles):
    filename = upload.name.strip().lower()

    queryset = AttachmentType.objects.filter(status=AttachmentType.ENABLED)
    for filetype in queryset.prefetch_related("limit_uploads_to"):
        for extension in filetype.extensions_list:
            if filename.endswith(".%s" % extension):
                break
        else:
            continue

        if (
            filetype.mimetypes_list
            and upload.content_type not in filetype.mimetypes_list
        ):
            continue

        if filetype.limit_uploads_to.exists():
            allowed_roles = set(r.pk for r in filetype.limit_uploads_to.all())
            if not user_roles & allowed_roles:
                continue

        return filetype

    raise ValidationError(
        pgettext("attachments api", "You can't upload files of this type.")
    )


def validate_filesize(upload, upload_limit):
    if upload.size > upload_limit * 1024:
        message = pgettext(
            "attachments api",
            "You can't upload files larger than %(limit)s (your file has %(upload)s).",
        )
        raise ValidationError(
            message
            % {
                "upload": filesizeformat(upload.size).rstrip(".0"),
                "limit": filesizeformat(upload_limit * 1024).rstrip(".0"),
            }
        )


def is_upload_image(upload):
    filename = upload.name.strip().lower()

    for extension in IMAGE_EXTENSIONS:
        if filename.endswith(".%s" % extension):
            return True
    return False



import pytest

from ...notifications.threads import ThreadNotifications


@pytest.fixture
def api_link(user):
    return f"/api/users/{user.id}/forum-options/"


def test_change_user_options_api_returns_error_for_guests(db, client, api_link):
    response = client.post(api_link)
    assert response.status_code == 403


def test_change_user_options_api_returns_error_for_empty_post(user_client, api_link):
    response = user_client.post(api_link)

    assert response.status_code == 400
    assert response.json() == {
        "limits_private_thread_invites_to": ["This field is required."],
        "watch_started_threads": ["This field is required."],
        "watch_replied_threads": ["This field is required."],
        "watch_new_private_threads_by_followed": ["This field is required."],
        "watch_new_private_threads_by_other_users": ["This field is required."],
        "notify_new_private_threads_by_followed": ["This field is required."],
        "notify_new_private_threads_by_other_users": ["This field is required."],
    }


def test_change_user_options_api_returns_error_for_invalid_data(user_client, api_link):
    response = user_client.post(
        api_link,
        data={
            "limits_private_thread_invites_to": 100,
            "watch_started_threads": 200,
            "watch_replied_threads": 300,
            "watch_new_private_threads_by_followed": 400,
            "watch_new_private_threads_by_other_users": 500,
            "notify_new_private_threads_by_followed": 600,
            "notify_new_private_threads_by_other_users": 700,
        },
    )

    assert response.status_code == 400
    assert response.json() == {
        "limits_private_thread_invites_to": ['"100" is not a valid choice.'],
        "watch_started_threads": ['"200" is not a valid choice.'],
        "watch_replied_threads": ['"300" is not a valid choice.'],
        "watch_new_private_threads_by_followed": ['"400" is not a valid choice.'],
        "watch_new_private_threads_by_other_users": ['"500" is not a valid choice.'],
        "notify_new_private_threads_by_followed": ['"600" is not a valid choice.'],
        "notify_new_private_threads_by_other_users": ['"700" is not a valid choice.'],
    }


def test_change_user_options_api_changes_user_options(user_client, api_link, user):
    NOTIFY_NONE = ThreadNotifications.NONE
    NOTIFY_SITE_ONLY = ThreadNotifications.SITE_ONLY
    NOTIFY_SITE_AND_EMAIL = ThreadNotifications.SITE_AND_EMAIL

    response = user_client.post(
        api_link,
        data={
            "limits_private_thread_invites_to": 0,
            "watch_started_threads": NOTIFY_NONE,
            "watch_replied_threads": NOTIFY_NONE,
            "watch_new_private_threads_by_followed": NOTIFY_NONE,
            "watch_new_private_threads_by_other_users": NOTIFY_NONE,
            "notify_new_private_threads_by_followed": NOTIFY_NONE,
            "notify_new_private_threads_by_other_users": NOTIFY_NONE,
        },
    )
    assert response.status_code == 200

    user.refresh_from_db()

    assert not user.is_hiding_presence
    assert user.limits_private_thread_invites_to == 0
    assert user.watch_started_threads == NOTIFY_NONE
    assert user.watch_replied_threads == NOTIFY_NONE
    assert user.watch_new_private_threads_by_followed == NOTIFY_NONE
    assert user.watch_new_private_threads_by_other_users == NOTIFY_NONE
    assert user.notify_new_private_threads_by_followed == NOTIFY_NONE
    assert user.notify_new_private_threads_by_other_users == NOTIFY_NONE

    response = user_client.post(
        api_link,
        data={
            "is_hiding_presence": False,
            "limits_private_thread_invites_to": 1,
            "watch_started_threads": NOTIFY_SITE_ONLY,
            "watch_replied_threads": NOTIFY_SITE_ONLY,
            "watch_new_private_threads_by_followed": NOTIFY_SITE_ONLY,
            "watch_new_private_threads_by_other_users": NOTIFY_SITE_ONLY,
            "notify_new_private_threads_by_followed": NOTIFY_SITE_ONLY,
            "notify_new_private_threads_by_other_users": NOTIFY_SITE_ONLY,
        },
    )
    assert response.status_code == 200

    user.refresh_from_db()

    assert not user.is_hiding_presence
    assert user.limits_private_thread_invites_to == 1
    assert user.watch_started_threads == NOTIFY_SITE_ONLY
    assert user.watch_replied_threads == NOTIFY_SITE_ONLY
    assert user.watch_new_private_threads_by_followed == NOTIFY_SITE_ONLY
    assert user.watch_new_private_threads_by_other_users == NOTIFY_SITE_ONLY
    assert user.notify_new_private_threads_by_followed == NOTIFY_SITE_ONLY
    assert user.notify_new_private_threads_by_other_users == NOTIFY_SITE_ONLY

    response = user_client.post(
        api_link,
        data={
            "is_hiding_presence": True,
            "limits_private_thread_invites_to": 2,
            "watch_started_threads": NOTIFY_SITE_AND_EMAIL,
            "watch_replied_threads": NOTIFY_SITE_AND_EMAIL,
            "watch_new_private_threads_by_followed": NOTIFY_SITE_AND_EMAIL,
            "watch_new_private_threads_by_other_users": NOTIFY_SITE_AND_EMAIL,
            "notify_new_private_threads_by_followed": NOTIFY_SITE_AND_EMAIL,
            "notify_new_private_threads_by_other_users": NOTIFY_SITE_AND_EMAIL,
        },
    )
    assert response.status_code == 200

    user.refresh_from_db()

    assert user.is_hiding_presence
    assert user.limits_private_thread_invites_to == 2
    assert user.watch_started_threads == NOTIFY_SITE_AND_EMAIL
    assert user.watch_replied_threads == NOTIFY_SITE_AND_EMAIL
    assert user.watch_new_private_threads_by_followed == NOTIFY_SITE_AND_EMAIL
    assert user.watch_new_private_threads_by_other_users == NOTIFY_SITE_AND_EMAIL
    assert user.notify_new_private_threads_by_followed == NOTIFY_SITE_AND_EMAIL
    assert user.notify_new_private_threads_by_other_users == NOTIFY_SITE_AND_EMAIL



from django.urls import path

from ...conf import settings

from ..views.attachment import attachment_server
from ..views.goto import (
    ThreadGotoPostView,
    ThreadGotoLastView,
    ThreadGotoNewView,
    ThreadGotoBestAnswerView,
    ThreadGotoUnapprovedView,
    PrivateThreadGotoPostView,
    PrivateThreadGotoLastView,
    PrivateThreadGotoNewView,
)
from ..views.list import ForumThreadsList, CategoryThreadsList, PrivateThreadsList
from ..views.subscribed import redirect_subscribed_to_watched
from ..views.thread import ThreadView, PrivateThreadView

LISTS_TYPES = ("all", "my", "new", "unread", "watched", "unapproved")


def threads_list_patterns(prefix, view, patterns):
    urls = []
    for i, pattern in enumerate(patterns):
        if i > 0:
            url_name = "%s-%s" % (prefix, LISTS_TYPES[i])
        else:
            url_name = prefix

        urls.append(
            path(
                pattern,
                view.as_view(),
                name=url_name,
                kwargs={"list_type": LISTS_TYPES[i]},
            )
        )

    return urls


if settings.MISAGO_THREADS_ON_INDEX:
    urlpatterns = threads_list_patterns(
        "threads",
        ForumThreadsList,
        (
            "",
            "my/",
            "new/",
            "unread/",
            "watched/",
            "unapproved/",
        ),
    )
else:
    urlpatterns = threads_list_patterns(
        "threads",
        ForumThreadsList,
        (
            "threads/",
            "threads/my/",
            "threads/new/",
            "threads/unread/",
            "threads/watched/",
            "threads/unapproved/",
        ),
    )

urlpatterns += threads_list_patterns(
    "category",
    CategoryThreadsList,
    (
        "c/<slug:slug>/<int:pk>/",
        "c/<slug:slug>/<int:pk>/my/",
        "c/<slug:slug>/<int:pk>/new/",
        "c/<slug:slug>/<int:pk>/unread/",
        "c/<slug:slug>/<int:pk>/watched/",
        "c/<slug:slug>/<int:pk>/unapproved/",
    ),
)

urlpatterns += threads_list_patterns(
    "private-threads",
    PrivateThreadsList,
    (
        "private-threads/",
        "private-threads/my/",
        "private-threads/new/",
        "private-threads/unread/",
        "private-threads/watched/",
    ),
)


# Redirect from subscribed to watched
if settings.MISAGO_THREADS_ON_INDEX:
    root_subscribed_path = "subscribed/"
else:
    root_subscribed_path = "threads/subscribed/"

urlpatterns += [
    path(root_subscribed_path, redirect_subscribed_to_watched),
    path("c/<slug:slug>/<int:pk>/subscribed/", redirect_subscribed_to_watched),
    path("private-threads/subscribed/", redirect_subscribed_to_watched),
]


def thread_view_patterns(prefix, view):
    urls = [
        path(
            "%s/<slug:slug>/<int:pk>/" % prefix[0],
            view.as_view(),
            name=prefix,
        ),
        path(
            "%s/<slug:slug>/<int:pk>/<int:page>/" % prefix[0],
            view.as_view(),
            name=prefix,
        ),
    ]
    return urls


urlpatterns += thread_view_patterns("thread", ThreadView)
urlpatterns += thread_view_patterns("private-thread", PrivateThreadView)


def goto_patterns(prefix, **views):
    urls = []

    post_view = views.pop("post", None)
    if post_view:
        url_pattern = "%s/<slug:slug>/<int:pk>/post/<int:post>/" % prefix[0]
        url_name = "%s-post" % prefix
        urls.append(path(url_pattern, post_view.as_view(), name=url_name))

    for name, view in views.items():
        name = name.replace("_", "-")
        url_pattern = "%s/<slug:slug>/<int:pk>/%s/" % (
            prefix[0],
            name,
        )
        url_name = "%s-%s" % (prefix, name)
        urls.append(path(url_pattern, view.as_view(), name=url_name))

    return urls


urlpatterns += goto_patterns(
    "thread",
    post=ThreadGotoPostView,
    last=ThreadGotoLastView,
    new=ThreadGotoNewView,
    best_answer=ThreadGotoBestAnswerView,
    unapproved=ThreadGotoUnapprovedView,
)

urlpatterns += goto_patterns(
    "private-thread",
    post=PrivateThreadGotoPostView,
    last=PrivateThreadGotoLastView,
    new=PrivateThreadGotoNewView,
)

urlpatterns += [
    path(
        "a/<slug:secret>/<int:pk>/",
        attachment_server,
        name="attachment",
    ),
    path(
        "a/thumb/<slug:secret>/<int:pk>/",
        attachment_server,
        name="attachment-thumbnail",
        kwargs={"thumbnail": True},
    ),
]



from dataclasses import dataclass

from django.contrib.auth import get_user_model
from django.forms import ValidationError
from django.utils.crypto import get_random_string
from unidecode import unidecode

from ..users.validators import (
    dj_validate_email,
    validate_username_content,
    validate_username_length,
)
from .exceptions import OAuth2UserDataValidationError
from .hooks import filter_user_data_hook, validate_user_data_hook

User = get_user_model()


class UsernameSettings:
    username_length_max: int = 200
    username_length_min: int = 1


def validate_user_data(request, user, user_data, response_json):
    try:
        return validate_user_data_hook(
            validate_user_data_action,
            request,
            user,
            user_data,
            response_json,
        )
    except ValidationError as exc:
        raise OAuth2UserDataValidationError(error_list=[str(exc.message)])


def validate_user_data_action(request, user, user_data, response_json):
    filtered_data = filter_user_data(request, user, user_data)
    validate_username_content(filtered_data["name"])
    validate_username_length(UsernameSettings, filtered_data["name"])
    dj_validate_email(filtered_data["email"])
    return filtered_data


def filter_user_data(request, user, user_data):
    return filter_user_data_hook(filter_user_data_action, request, user, user_data)


def filter_user_data_action(request, user, user_data):
    return {
        "id": user_data["id"],
        "name": filter_name(user, user_data["name"] or ""),
        "email": user_data["email"] or "",
        "avatar": user_data["avatar"],
    }


def filter_user_data_with_filters(request, user, user_data, filters):
    for filter_user_data in filters:
        user_data = filter_user_data(request, user, user_data.copy()) or user_data
    return user_data


def filter_name(user, name):
    if user and user.username == name:
        return name

    clean_name = "".join(
        [c for c in unidecode(name.replace(" ", "_")) if c.isalnum() or c == "_"]
    )

    if user and user.username == clean_name:
        return clean_name  # No change in name

    if not clean_name.replace("_", ""):
        clean_name = "User_%s" % get_random_string(4)

    clean_name_root = clean_name
    while True:
        try:
            db_user = User.objects.get_by_username(clean_name)
        except User.DoesNotExist:
            return clean_name

        if not user or user.pk != db_user.pk:
            clean_name = f"{clean_name_root}_{get_random_string(4)}"
        else:
            return clean_name



from .. import test
from ...users.test import create_test_user
from ..models import ThreadParticipant
from .test_privatethreads import PrivateThreadsTestCase


class SyncUnreadPrivateThreadsTestCase(PrivateThreadsTestCase):
    def setUp(self):
        super().setUp()

        self.other_user = create_test_user("Other_User", "user@example.com")
        self.thread = test.post_thread(self.category, poster=self.user)

        ThreadParticipant.objects.set_owner(self.thread, self.other_user)
        ThreadParticipant.objects.add_participants(self.thread, [self.user])

    def test_middleware_counts_new_thread(self):
        """middleware counts new thread"""
        self.user.sync_unread_private_threads = True
        self.user.save()

        response = self.client.get("/")
        self.assertEqual(response.status_code, 200)

        # user was resynced
        self.reload_user()

        self.assertFalse(self.user.sync_unread_private_threads)
        self.assertEqual(self.user.unread_private_threads, 1)

    def test_middleware_counts_unread_thread(self):
        """
        middleware counts thread with unread reply, post read flags user for recount
        """
        self.user.sync_unread_private_threads = True
        self.user.save()

        self.client.post(self.thread.last_post.get_read_api_url())

        # post read zeroed list of unread private threads
        self.reload_user()
        self.assertFalse(self.user.sync_unread_private_threads)
        self.assertEqual(self.user.unread_private_threads, 0)

        # reply to thread
        test.reply_thread(self.thread)

        self.user.sync_unread_private_threads = True
        self.user.save()

        # middleware did recount and accounted for new unread post
        response = self.client.get("/")
        self.assertEqual(response.status_code, 200)

        self.reload_user()
        self.assertFalse(self.user.sync_unread_private_threads)
        self.assertEqual(self.user.unread_private_threads, 1)



from ..models import WatchedThread
from ..threads import get_watched_thread


def test_get_watched_thread_returns_null_if_thread_is_not_watched(user, thread):
    watched_thread = get_watched_thread(user, thread)
    assert watched_thread is None


def test_get_watched_thread_returns_watched_thread_for_user(
    user, other_user, thread, watched_thread_factory
):
    watched_thread_factory(other_user, thread, send_emails=True)
    user_watched_thread = watched_thread_factory(user, thread, send_emails=True)

    watched_thread = get_watched_thread(user, thread)
    assert watched_thread == user_watched_thread


def test_get_watched_thread_returns_watched_thread_for_user_and_thread(
    user, thread, other_thread, watched_thread_factory
):
    watched_thread_factory(user, other_thread, send_emails=True)

    watched_thread = get_watched_thread(user, thread)
    assert watched_thread is None


def test_get_watched_thread_returns_first_watched_thread_for_user_if_multiple_exist(
    user, thread, watched_thread_factory
):
    first_watched_thread = watched_thread_factory(user, thread, send_emails=True)
    watched_thread_factory(user, thread, send_emails=False)

    watched_thread = get_watched_thread(user, thread)
    assert watched_thread == first_watched_thread


def test_get_watched_thread_removes_extra_watched_threads_for_user(
    user, thread, watched_thread_factory
):
    first_watched_thread = watched_thread_factory(user, thread, send_emails=True)
    watched_thread_factory(user, thread, send_emails=False)

    assert WatchedThread.objects.count() == 2

    watched_thread = get_watched_thread(user, thread)
    assert watched_thread == first_watched_thread

    assert WatchedThread.objects.count() == 1



import pytest

from ..hooks import FilterHook

ACTION = 0
FIRST_FILTER = 1
SECOND_FILTER = 2


class MockFilterHook(FilterHook):
    def __call__(self, action):
        return super().__call__(action, [])


def action(data):
    return [ACTION]


def first_filter(action, data):
    return [action(data), FIRST_FILTER]


def second_filter(action, data):
    data = action(data)
    return [action(data), SECOND_FILTER]


@pytest.fixture
def hook():
    return MockFilterHook()


def test_filter_hook_without_filters_just_calls_action(hook):
    assert hook(action) == [ACTION]


def test_filter_hook_without_actions_is_falsy(hook):
    assert not hook


def test_filter_hook_with_actions_is_truthy(hook):
    hook.append_filter(first_filter)
    assert hook


def test_filter_hook_calls_filter_before_action(hook):
    hook.append_filter(first_filter)
    assert hook(action) == [[ACTION], FIRST_FILTER]


def test_filter_hook_calls_filters_in_order_of_adding(hook):
    hook.append_filter(first_filter)
    hook.append_filter(second_filter)
    assert hook(action) == [[[ACTION], FIRST_FILTER], SECOND_FILTER]


def test_filter_can_be_inserted_before_other_filters(hook):
    hook.append_filter(first_filter)
    hook.prepend_filter(second_filter)
    assert hook(action) == [[[ACTION], SECOND_FILTER], FIRST_FILTER]



from .cache import get_settings_cache, set_settings_cache
from .models import Setting


class DynamicSettings:
    _overrides = {}

    def __init__(self, cache_versions):
        self._settings = get_settings_cache(cache_versions)
        if self._settings is None:
            self._settings = get_settings_from_db()
            set_settings_cache(cache_versions, self._settings)

    def get(self, setting):
        return self._settings.get(setting)

    def get_public_settings(self):
        public_settings = {}
        for name, setting in self._settings.items():
            if setting["is_public"]:
                public_settings[name] = setting["value"]
        return public_settings

    def get_lazy_setting_value(self, setting):
        try:
            if self._settings[setting]["is_lazy"]:
                if setting in self._overrides:
                    return self._overrides[setting]
                if not self._settings[setting].get("real_value"):
                    real_value = Setting.objects.get(setting=setting).value
                    self._settings[setting]["real_value"] = real_value
                return self._settings[setting]["real_value"]
            raise ValueError("Setting %s is not lazy" % setting)
        except (KeyError, Setting.DoesNotExist):
            raise AttributeError("Setting %s is not defined" % setting)

    def __getattr__(self, setting):
        if setting in self._overrides:
            return self._overrides[setting]
        if setting in self._settings:
            return self._settings[setting]["value"]
        raise AttributeError("Setting %s is not defined" % setting)

    @classmethod
    def override_settings(cls, overrides):
        cls._overrides = overrides

    @classmethod
    def remove_overrides(cls):
        cls._overrides = {}


def get_settings_from_db():
    settings = {}
    for setting in Setting.objects.iterator():
        settings[setting.setting] = {
            "value": None,
            "is_lazy": setting.is_lazy,
            "is_public": setting.is_public,
            "width": None,
            "height": None,
        }

        if setting.is_lazy:
            settings[setting.setting]["value"] = True if setting.value else None
        elif setting.python_type == "image":
            settings[setting.setting].update(
                {
                    "value": setting.value.url if setting.value else None,
                    "width": setting.image_width,
                    "height": setting.image_height,
                }
            )
        else:
            settings[setting.setting]["value"] = setting.value

    return settings



from . import downloaded, dynamic, gallery, gravatar

SET_DEFAULT_AVATAR = {
    "gravatar": gravatar.set_avatar,
    "dynamic": dynamic.set_avatar,
    "gallery": gallery.set_random_avatar,
}


def set_default_avatar(user, default_avatar, gravatar_fallback):
    try:
        SET_DEFAULT_AVATAR[default_avatar](user)
    except RuntimeError:
        if gallery.galleries_exist():
            SET_DEFAULT_AVATAR[gravatar_fallback](user)
        else:
            dynamic.set_avatar(user)


def set_default_avatar_from_url(user, avatar_url):
    try:
        downloaded.set_avatar(user, avatar_url)
    except RuntimeError:
        dynamic.set_avatar(user)



from django import template

register = template.Library()


@register.filter
def batch(items, size):
    batch_size = 0
    batch_items = []

    for item in items:
        batch_size += 1
        batch_items.append(item)

        if batch_size == size:
            yield batch_items
            batch_size = 0
            batch_items = []

    if batch_items:
        yield batch_items


@register.filter
def batchnonefilled(items, size):
    batch_size = 0
    batch_items = []

    for item in items:
        batch_size += 1
        batch_items.append(item)

        if batch_size == size:
            yield batch_items
            batch_size = 0
            batch_items = []

    if batch_size:
        batch_items.extend([None] * (size - batch_size))
        yield batch_items



from django.urls import reverse

from ...test import assert_contains


def test_custom_menu_items_are_rendered_in_navbar(client, navar_menu_item):
    response = client.get(reverse("misago:index"))
    assert_contains(response, navar_menu_item.url)


def test_custom_menu_items_are_rendered_in_footer(client, footer_menu_item):
    response = client.get(reverse("misago:index"))
    assert_contains(response, footer_menu_item.url)


def test_custom_menu_items_are_rendered_in_both_menus(client, both_menus_item):
    response = client.get(reverse("misago:index"))
    assert_contains(response, both_menus_item.url)



"""Misago's default Django Project settings."""

__all__ = [
    "INSTALLED_APPS",
    "INSTALLED_PLUGINS",
    "MISAGO_NOTIFICATIONS_RETRY_DELAY",
    "TEMPLATE_CONTEXT_PROCESSORS",
]

INSTALLED_APPS = [
    # Misago overrides for Django core feature
    "misago",
    "misago.users",
    # Django apps
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.postgres",
    "django.contrib.humanize",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    # 3rd party apps used by Misago
    "ariadne_django",
    "celery",
    "debug_toolbar",
    "mptt",
    "rest_framework",
    "social_django",
    # Misago apps
    "misago.admin",
    "misago.acl",
    "misago.analytics",
    "misago.cache",
    "misago.core",
    "misago.conf",
    "misago.icons",
    "misago.themes",
    "misago.markup",
    "misago.legal",
    "misago.notifications",
    "misago.categories",
    "misago.threads",
    "misago.readtracker",
    "misago.search",
    "misago.oauth2",
    "misago.socialauth",
    "misago.graphql",
    "misago.faker",
    "misago.menus",
    "misago.plugins",
    "misago.apiv2",
]

INSTALLED_PLUGINS = []

TEMPLATE_CONTEXT_PROCESSORS = [
    "django.template.context_processors.debug",
    "django.template.context_processors.i18n",
    "django.template.context_processors.media",
    "django.template.context_processors.request",
    "django.template.context_processors.static",
    "django.template.context_processors.tz",
    "django.contrib.auth.context_processors.auth",
    "django.contrib.messages.context_processors.messages",
    "misago.acl.context_processors.user_acl",
    "misago.conf.context_processors.conf",
    "misago.conf.context_processors.og_image",
    "misago.core.context_processors.misago_version",
    "misago.core.context_processors.request_path",
    "misago.core.context_processors.momentjs_locale",
    "misago.icons.context_processors.icons",
    "misago.search.context_processors.search_providers",
    "misago.themes.context_processors.theme",
    "misago.legal.context_processors.legal_links",
    "misago.menus.context_processors.menus",
    "misago.users.context_processors.user_links",
    "misago.core.context_processors.hooks",
    # Data preloaders
    "misago.conf.context_processors.preload_settings_json",
    "misago.core.context_processors.current_link",
    "misago.markup.context_processors.preload_api_url",
    "misago.threads.context_processors.preload_threads_urls",
    "misago.users.context_processors.preload_user_json",
    "misago.categories.context_processors.preload_categories_json",
    "misago.socialauth.context_processors.preload_socialauth_json",
    # Note: keep frontend_context processor last for previous processors
    # to be able to expose data UI app via request.frontend_context
    "misago.core.context_processors.frontend_context",
]

MISAGO_NOTIFICATIONS_RETRY_DELAY = 5  # Seconds



import pytest
from django.utils import timezone

from .acl import ACL_CACHE, useracl
from .admin.auth import authorize_admin
from .categories.models import Category
from .conf import SETTINGS_CACHE
from .conf.dynamicsettings import DynamicSettings
from .conf.staticsettings import StaticSettings
from .menus import MENU_ITEMS_CACHE
from .notifications.models import WatchedThread
from .notifications.threads import ThreadNotifications
from .socialauth import SOCIALAUTH_CACHE
from .test import MisagoClient
from .themes import THEME_CACHE
from .threads.models import Thread
from .threads.test import post_thread, reply_thread
from .users import BANS_CACHE
from .users.models import AnonymousUser
from .users.test import create_test_superuser, create_test_user


def get_cache_versions():
    return {
        ACL_CACHE: "abcdefgh",
        BANS_CACHE: "abcdefgh",
        SETTINGS_CACHE: "abcdefgh",
        SOCIALAUTH_CACHE: "abcdefgh",
        THEME_CACHE: "abcdefgh",
        MENU_ITEMS_CACHE: "abcdefgh",
    }


@pytest.fixture
def cache_versions():
    return get_cache_versions()


@pytest.fixture
def dynamic_settings(db, cache_versions):
    return DynamicSettings(cache_versions)


@pytest.fixture
def settings():
    return StaticSettings()


@pytest.fixture
def user_password():
    return "p4ssw0rd!"


@pytest.fixture
def anonymous_user():
    return AnonymousUser()


@pytest.fixture
def anonymous_user_acl(anonymous_user, cache_versions):
    return useracl.get_user_acl(anonymous_user, cache_versions)


@pytest.fixture
def user(db, user_password):
    return create_test_user("User", "user@example.com", user_password)


@pytest.fixture
def user_acl(user, cache_versions):
    return useracl.get_user_acl(user, cache_versions)


@pytest.fixture
def other_user(db, user_password):
    return create_test_user("Other_User", "otheruser@example.com", user_password)


@pytest.fixture
def other_user_acl(other_user, cache_versions):
    return useracl.get_user_acl(other_user, cache_versions)


@pytest.fixture
def staffuser(db, user_password):
    user = create_test_superuser("Staff_User", "staffuser@example.com", user_password)
    user.is_superuser = False
    user.save()
    return user


@pytest.fixture
def staffuser_acl(staffuser, cache_versions):
    return useracl.get_user_acl(staffuser, cache_versions)


@pytest.fixture
def other_staffuser(db, user_password):
    user = create_test_superuser(
        "Other_Staff_User", "otherstaffuser@example.com", user_password
    )

    user.is_superuser = False
    user.save()
    return user


@pytest.fixture
def superuser(db, user_password):
    return create_test_superuser("Super_User", "superuser@example.com", user_password)


@pytest.fixture
def superuser_acl(superuser, cache_versions):
    return useracl.get_user_acl(superuser, cache_versions)


@pytest.fixture
def other_superuser(db, user_password):
    return create_test_superuser(
        "OtherSuperUser", "othersuperuser@example.com", user_password
    )


@pytest.fixture
def inactive_user(db, user_password):
    return create_test_user(
        "Inactive_User", "inactiveuser@example.com", user_password, is_active=False
    )


@pytest.fixture
def client():
    return MisagoClient()


@pytest.fixture
def user_client(client, user):
    client.force_login(user)
    session = client.session
    session.save()
    return client


@pytest.fixture
def admin_client(mocker, client, superuser):
    client.force_login(superuser)
    session = client.session
    authorize_admin(mocker.Mock(session=session, user=superuser))
    session.save()
    return client


@pytest.fixture
def staff_client(mocker, client, staffuser):
    client.force_login(staffuser)
    session = client.session
    authorize_admin(mocker.Mock(session=session, user=staffuser))
    session.save()
    return client


@pytest.fixture
def root_category(db):
    return Category.objects.root_category()


@pytest.fixture
def private_threads_category(db):
    return Category.objects.private_threads()


@pytest.fixture
def default_category(db):
    return Category.objects.get(slug="first-category")


@pytest.fixture
def thread(default_category):
    return post_thread(default_category)


@pytest.fixture
def other_thread(default_category):
    return post_thread(default_category)


@pytest.fixture
def hidden_thread(default_category):
    return post_thread(default_category, is_hidden=True)


@pytest.fixture
def unapproved_thread(default_category):
    return post_thread(default_category, is_unapproved=True)


@pytest.fixture
def post(thread):
    return thread.first_post


@pytest.fixture
def reply(thread):
    return reply_thread(thread, poster="Ghost", posted_on=timezone.now())


@pytest.fixture
def user_reply(thread, user):
    return reply_thread(thread, poster=user, posted_on=timezone.now())


@pytest.fixture
def other_user_reply(thread, other_user):
    return reply_thread(thread, poster=other_user, posted_on=timezone.now())


@pytest.fixture
def user_thread(default_category, user):
    return post_thread(default_category, poster=user)


@pytest.fixture
def user_hidden_thread(default_category, user):
    return post_thread(default_category, poster=user, is_hidden=True)


@pytest.fixture
def user_unapproved_thread(default_category, user):
    return post_thread(default_category, poster=user, is_unapproved=True)


@pytest.fixture
def other_user_thread(default_category, other_user):
    return post_thread(default_category, poster=other_user)


@pytest.fixture
def other_user_hidden_thread(default_category, other_user):
    return post_thread(default_category, poster=other_user, is_hidden=True)


@pytest.fixture
def other_user_unapproved_thread(default_category, other_user):
    return post_thread(default_category, poster=other_user, is_unapproved=True)


@pytest.fixture
def private_thread(private_threads_category):
    return post_thread(private_threads_category)


@pytest.fixture
def private_thread_reply(private_thread):
    return reply_thread(private_thread, poster="Ghost", posted_on=timezone.now())


@pytest.fixture
def private_thread_user_reply(private_thread, user):
    return reply_thread(private_thread, poster=user, posted_on=timezone.now())


@pytest.fixture
def user_private_thread(user, private_threads_category):
    return post_thread(private_threads_category, poster=user)


@pytest.fixture
def categories_tree(root_category):
    sibling_category = Category(
        name="Sibling Category",
        slug="sibling-category",
    )

    sibling_category.insert_at(root_category, position="last-child", save=True)

    child_category = Category(
        name="Child Category",
        slug="child-category",
        color="#FF0000",
    )

    child_category.insert_at(sibling_category, position="last-child", save=True)

    other_category = Category(
        name="Other Category",
        slug="other-category",
        short_name="Other",
    )

    other_category.insert_at(root_category, position="last-child", save=True)


@pytest.fixture
def sibling_category(categories_tree):
    return Category.objects.get(slug="sibling-category")


@pytest.fixture
def child_category(categories_tree):
    return Category.objects.get(slug="child-category")


@pytest.fixture
def other_category(categories_tree):
    return Category.objects.get(slug="other-category")


@pytest.fixture
def watched_thread_factory():
    def create_watched_thread(user: "User", thread: "Thread", send_emails: bool):
        return WatchedThread.objects.create(
            user=user,
            category_id=thread.category_id,
            thread=thread,
            send_emails=send_emails,
        )

    return create_watched_thread



from django.urls import path

from .views import google_site_verification

urlpatterns = [
    path(
        "google<slug:token>.html",
        google_site_verification,
        name="google-site-verification",
    )
]



from urllib.parse import urlparse

from django.urls import Resolver404, resolve

from .models import PostLike


def add_categories_to_items(root_category, categories, items):
    categories_dict = {}
    for category in categories:
        categories_dict[category.pk] = category
        if category.parent_id:
            category.parent = categories_dict[category.parent_id]

    for item in items:
        item.category = categories_dict[item.category_id]


def add_likes_to_posts(user, posts):
    if user.is_anonymous:
        return

    posts_map = {}
    for post in posts:
        posts_map[post.id] = post
        post.is_liked = False

    queryset = PostLike.objects.filter(liker=user, post_id__in=posts_map.keys())

    for like in queryset.values("post_id"):
        posts_map[like["post_id"]].is_liked = True


SUPPORTED_THREAD_ROUTES = {
    "misago:thread": "pk",
    "misago:thread-post": "pk",
    "misago:thread-last": "pk",
    "misago:thread-new": "pk",
    "misago:thread-unapproved": "pk",
}


def get_thread_id_from_url(request, url):  # pylint: disable=too-many-return-statements
    clean_url = str(url).strip()
    url_bits = urlparse(clean_url)

    if url_bits.netloc and url_bits.netloc != request.get_host():
        return None

    if url_bits.path.startswith(request.get_host()):
        clean_path = url_bits.path.lstrip(request.get_host())
    else:
        clean_path = url_bits.path

    wsgi_alias = request.path[: len(request.path_info) * -1]
    if wsgi_alias and not clean_path.startswith(wsgi_alias):
        return None

    try:
        resolution = resolve(clean_path[len(wsgi_alias) :])
    except Resolver404:
        return None

    if not resolution.namespaces:
        return None

    url_name = "%s:%s" % (":".join(resolution.namespaces), resolution.url_name)
    kwargname = SUPPORTED_THREAD_ROUTES.get(url_name)

    if not kwargname:
        return None

    try:
        return int(resolution.kwargs.get(kwargname))
    except (TypeError, ValueError):
        return None



# Generated by Django 3.2.15 on 2022-11-01 21:11

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("misago_categories", "0008_auto_20190518_1659"),
    ]

    operations = [
        migrations.AddField(
            model_name="category",
            name="color",
            field=models.CharField(blank=True, max_length=7, null=True),
        ),
        migrations.AddField(
            model_name="category",
            name="short_name",
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]



from .followers import Followers


class Follows(Followers):
    def get_queryset(self, profile):
        return profile.follows

    def get_template_context(self):
        return {"follows": self.users, "count": self.paginator["count"]}



from unittest.mock import Mock

import pytest
from rest_framework import serializers

from .. import test
from ...conf.test import override_dynamic_settings
from ..api.postingendpoint import PostingEndpoint
from ..api.postingendpoint.attachments import (
    AttachmentsMiddleware,
    validate_attachments_count,
)
from ..models import Attachment, AttachmentType


@pytest.fixture
def context(default_category, dynamic_settings, user, user_acl):
    thread = test.post_thread(category=default_category)
    post = thread.first_post
    post.update_fields = []

    return {
        "category": default_category,
        "thread": thread,
        "post": post,
        "settings": dynamic_settings,
        "user": user,
        "user_acl": user_acl,
    }


def create_attachment(*, post=None, user=None):
    return Attachment.objects.create(
        secret=Attachment.generate_new_secret(),
        filetype=AttachmentType.objects.order_by("id").last(),
        post=post,
        size=1000,
        uploader=user if user else None,
        uploader_name=user.username if user else "testuser",
        uploader_slug=user.slug if user else "testuser",
        filename="testfile_%s.zip" % (Attachment.objects.count() + 1),
    )


def test_middleware_is_used_if_user_has_permission_to_upload_attachments(context):
    context["user_acl"]["max_attachment_size"] = 1024
    middleware = AttachmentsMiddleware(**context)
    assert middleware.use_this_middleware()


def test_middleware_is_not_used_if_user_has_no_permission_to_upload_attachments(
    context,
):
    context["user_acl"]["max_attachment_size"] = 0
    middleware = AttachmentsMiddleware(**context)
    assert not middleware.use_this_middleware()


def test_middleware_handles_no_data(context):
    middleware = AttachmentsMiddleware(
        request=Mock(data={}), mode=PostingEndpoint.START, **context
    )

    serializer = middleware.get_serializer()
    assert serializer.is_valid()


def test_middleware_handles_empty_data(context):
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": []}), mode=PostingEndpoint.START, **context
    )

    serializer = middleware.get_serializer()
    assert serializer.is_valid()


def test_data_validation_fails_if_attachments_data_is_not_iterable(context):
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": "none"}),
        mode=PostingEndpoint.START,
        **context
    )

    serializer = middleware.get_serializer()
    assert not serializer.is_valid()


def test_data_validation_fails_if_attachments_data_has_non_int_values(context):
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": [1, "b"]}),
        mode=PostingEndpoint.START,
        **context
    )

    serializer = middleware.get_serializer()
    assert not serializer.is_valid()


@override_dynamic_settings(post_attachments_limit=2)
def test_data_validation_fails_if_attachments_data_is_longer_than_allowed(context):
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": range(5)}),
        mode=PostingEndpoint.START,
        **context
    )

    serializer = middleware.get_serializer()
    assert not serializer.is_valid()


def test_middleware_adds_attachment_to_new_post(context):
    new_attachment = create_attachment(user=context["user"])
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": [new_attachment.id]}),
        mode=PostingEndpoint.START,
        **context
    )
    serializer = middleware.get_serializer()
    serializer.is_valid()
    middleware.save(serializer)

    new_attachment.refresh_from_db()
    assert new_attachment.post == context["post"]


def test_middleware_adds_attachment_to_attachments_cache(context):
    new_attachment = create_attachment(user=context["user"])
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": [new_attachment.id]}),
        mode=PostingEndpoint.START,
        **context
    )
    serializer = middleware.get_serializer()
    serializer.is_valid()
    middleware.save(serializer)

    attachments_cache = context["post"].attachments_cache
    assert len(attachments_cache) == 1
    assert attachments_cache[0]["id"] == new_attachment.id


def test_middleware_adds_attachment_to_existing_post(context):
    new_attachment = create_attachment(user=context["user"])
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": [new_attachment.id]}),
        mode=PostingEndpoint.EDIT,
        **context
    )
    serializer = middleware.get_serializer()
    serializer.is_valid()
    middleware.save(serializer)

    new_attachment.refresh_from_db()
    assert new_attachment.post == context["post"]


def test_middleware_adds_attachment_to_post_with_existing_attachment(context):
    old_attachment = create_attachment(post=context["post"])
    new_attachment = create_attachment(user=context["user"])
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": [old_attachment.id, new_attachment.id]}),
        mode=PostingEndpoint.EDIT,
        **context
    )
    serializer = middleware.get_serializer()
    serializer.is_valid()
    middleware.save(serializer)

    new_attachment.refresh_from_db()
    assert new_attachment.post == context["post"]

    old_attachment.refresh_from_db()
    assert old_attachment.post == context["post"]


def test_middleware_adds_attachment_to_existing_attachments_cache(context):
    old_attachment = create_attachment(post=context["post"])
    new_attachment = create_attachment(user=context["user"])
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": [old_attachment.id, new_attachment.id]}),
        mode=PostingEndpoint.EDIT,
        **context
    )
    serializer = middleware.get_serializer()
    serializer.is_valid()
    middleware.save(serializer)

    attachments_cache = context["post"].attachments_cache
    assert len(attachments_cache) == 2
    assert attachments_cache[0]["id"] == new_attachment.id
    assert attachments_cache[1]["id"] == old_attachment.id


def test_other_user_attachment_cant_be_added_to_post(context):
    attachment = create_attachment()
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": [attachment.id]}),
        mode=PostingEndpoint.EDIT,
        **context
    )
    serializer = middleware.get_serializer()
    serializer.is_valid()
    middleware.save(serializer)

    attachment.refresh_from_db()
    assert not attachment.post


def test_other_post_attachment_cant_be_added_to_new_post(context, default_category):
    post = test.post_thread(category=default_category).first_post
    attachment = create_attachment(post=post, user=context["user"])
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": [attachment.id]}),
        mode=PostingEndpoint.EDIT,
        **context
    )
    serializer = middleware.get_serializer()
    serializer.is_valid()
    middleware.save(serializer)

    attachment.refresh_from_db()
    assert attachment.post == post


def test_middleware_removes_attachment_from_post(context):
    attachment = create_attachment(post=context["post"], user=context["user"])
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": []}), mode=PostingEndpoint.EDIT, **context
    )
    serializer = middleware.get_serializer()
    serializer.is_valid()
    middleware.save(serializer)

    context["post"].refresh_from_db()
    assert not context["post"].attachment_set.exists()


def test_middleware_removes_attachment_from_attachments_cache(context):
    attachment = create_attachment(post=context["post"], user=context["user"])
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": []}), mode=PostingEndpoint.EDIT, **context
    )
    serializer = middleware.get_serializer()
    serializer.is_valid()
    middleware.save(serializer)

    assert not context["post"].attachments_cache


def test_middleware_deletes_attachment_removed_from_post(context):
    attachment = create_attachment(post=context["post"], user=context["user"])
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": []}), mode=PostingEndpoint.EDIT, **context
    )
    serializer = middleware.get_serializer()
    serializer.is_valid()
    middleware.save(serializer)

    with pytest.raises(Attachment.DoesNotExist):
        attachment.refresh_from_db()


def test_middleware_blocks_user_from_removing_other_user_attachment_without_permission(
    context,
):
    attachment = create_attachment(post=context["post"])
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": []}), mode=PostingEndpoint.EDIT, **context
    )
    serializer = middleware.get_serializer()
    assert not serializer.is_valid()
    middleware.save(serializer)

    attachment.refresh_from_db()
    assert attachment.post == context["post"]


def test_middleware_allows_user_with_permission_to_remove_other_user_attachment(
    context,
):
    context["user_acl"]["can_delete_other_users_attachments"] = True
    attachment = create_attachment(post=context["post"])
    middleware = AttachmentsMiddleware(
        request=Mock(data={"attachments": []}), mode=PostingEndpoint.EDIT, **context
    )
    serializer = middleware.get_serializer()
    serializer.is_valid()
    middleware.save(serializer)

    context["post"].refresh_from_db()
    assert not context["post"].attachment_set.exists()


def test_attachments_count_validator_allows_attachments_within_limit():
    settings = Mock(post_attachments_limit=5)
    validate_attachments_count(range(5), settings)


def test_attachments_count_validator_raises_validation_error_on_too_many_attachmes():
    settings = Mock(post_attachments_limit=2)
    with pytest.raises(serializers.ValidationError):
        validate_attachments_count(range(5), settings)



from django.urls import reverse
from rest_framework import serializers

from ...core.serializers import MutableFields
from ...users.serializers import UserSerializer as BaseUserSerializer
from ..models import Post

__all__ = ["PostSerializer"]

UserSerializer = BaseUserSerializer.subset_fields(
    "id",
    "username",
    "real_name",
    "rank",
    "avatars",
    "signature",
    "title",
    "status",
    "posts",
    "url",
)


class PostSerializer(serializers.ModelSerializer, MutableFields):
    poster = UserSerializer(many=False, read_only=True)
    content = serializers.SerializerMethodField()
    attachments = serializers.SerializerMethodField()
    last_editor = serializers.PrimaryKeyRelatedField(read_only=True)
    hidden_by = serializers.PrimaryKeyRelatedField(read_only=True)

    acl = serializers.SerializerMethodField()
    is_read = serializers.SerializerMethodField()
    is_new = serializers.SerializerMethodField()
    is_liked = serializers.SerializerMethodField()
    is_protected = serializers.SerializerMethodField()
    last_likes = serializers.SerializerMethodField()
    likes = serializers.SerializerMethodField()

    api = serializers.SerializerMethodField()
    url = serializers.SerializerMethodField()

    class Meta:
        model = Post
        fields = [
            "id",
            "poster",
            "poster_name",
            "content",
            "attachments",
            "posted_on",
            "updated_on",
            "hidden_on",
            "edits",
            "last_editor",
            "last_editor_name",
            "last_editor_slug",
            "hidden_by",
            "hidden_by_name",
            "hidden_by_slug",
            "is_unapproved",
            "is_hidden",
            "is_protected",
            "is_event",
            "event_type",
            "event_context",
            "acl",
            "is_liked",
            "is_new",
            "is_read",
            "last_likes",
            "likes",
            "api",
            "url",
        ]

    def get_content(self, obj):
        if (
            obj.is_valid
            and not obj.is_event
            and (not obj.is_hidden or obj.acl["can_see_hidden"])
        ):
            return obj.content

    def get_attachments(self, obj):
        return obj.attachments_cache

    def get_acl(self, obj):
        try:
            return obj.acl
        except AttributeError:
            return None

    def get_is_liked(self, obj):
        try:
            return obj.is_liked
        except AttributeError:
            return None

    def get_is_new(self, obj):
        try:
            return obj.is_new
        except AttributeError:
            return None

    def get_is_read(self, obj):
        try:
            return obj.is_read
        except AttributeError:
            return None

    def get_is_protected(self, obj):
        if obj.is_event:
            return None

        try:
            if obj.acl["can_see_protected"]:
                return obj.is_protected
        except AttributeError:
            pass

        return None

    def get_last_likes(self, obj):
        if obj.is_event:
            return None

        try:
            if obj.acl["can_see_likes"]:
                return obj.last_likes
        except AttributeError:
            return None

    def get_likes(self, obj):
        if obj.is_event:
            return None

        try:
            if obj.acl["can_see_likes"]:
                return obj.likes
        except AttributeError:
            return None

    def get_api(self, obj):
        api_links = {
            "index": obj.get_api_url(),
            "likes": obj.get_likes_api_url(),
            "editor": obj.get_editor_api_url(),
            "edits": obj.get_edits_api_url(),
            "read": obj.get_read_api_url(),
        }

        if obj.is_event:
            del api_links["likes"]

        return api_links

    def get_url(self, obj):
        return {
            "index": obj.get_absolute_url(),
            "last_editor": self.get_last_editor_url(obj),
            "hidden_by": self.get_hidden_by_url(obj),
        }

    def get_last_editor_url(self, obj):
        if obj.last_editor_id:
            return reverse(
                "misago:user",
                kwargs={"pk": obj.last_editor_id, "slug": obj.last_editor_slug},
            )

    def get_hidden_by_url(self, obj):
        if obj.hidden_by_id:
            return reverse(
                "misago:user",
                kwargs={"pk": obj.hidden_by_id, "slug": obj.hidden_by_slug},
            )



# Generated by Django 1.11.5 on 2017-10-15 20:10
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("misago_readtracker", "0003_migrate_reads_to_posts")]

    operations = [
        migrations.RemoveField(model_name="categoryread", name="category"),
        migrations.RemoveField(model_name="categoryread", name="user"),
        migrations.RemoveField(model_name="threadread", name="category"),
        migrations.RemoveField(model_name="threadread", name="thread"),
        migrations.RemoveField(model_name="threadread", name="user"),
        migrations.DeleteModel(name="CategoryRead"),
        migrations.DeleteModel(name="ThreadRead"),
    ]



from django.urls import reverse
from rest_framework import serializers

from ..models import PostEdit

__all__ = ["PostEditSerializer"]


class PostEditSerializer(serializers.ModelSerializer):
    diff = serializers.SerializerMethodField()

    url = serializers.SerializerMethodField()

    class Meta:
        model = PostEdit
        fields = ["id", "edited_on", "editor_name", "editor_slug", "diff", "url"]

    def get_diff(self, obj):
        return obj.get_diff()

    def get_url(self, obj):
        return {"editor": self.get_editor_url(obj)}

    def get_editor_url(self, obj):
        if obj.editor_id:
            return reverse(
                "misago:user", kwargs={"slug": obj.editor_slug, "pk": obj.editor_id}
            )



from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class MisagoPlugin:
    """
    A frozen dataclass with plugin's metadata.

    # Optional arguments

    ## `name: str`

    A string with the plugin name. Limited to 100 characters.

    ## `description: str`

    A string with the plugin description. Limited to 250 characters.

    ## `license: str`

    A string with the plugin license. Limited to 50 characters.

    ## `icon: str`

    A string with the plugin icon. Must be a valid Font Awesome icon CSS name,
    e.g., `fa fa-icon` or `fas fa-other-icon`.

    ## `color: str`

    A string with the plugin icon's color. Must be a color hex format prefixed
    with `#`, e.g., `#efefef`.

    ## `version: str`

    A string with the plugin version. Limited to 50 characters.

    ## `author: str`

    A string with the plugin author's name. Limited to 150 characters.

    ## `homepage: str`

    A string with the URL to the plugin's homepage.

    ## `sponsor: str`

    A string with the URL to a page with sponsorship instructions or a donation form.

    ## `help: str`

    A string with the URL to the plugin's help page or a support forum.

    ## `bugs: str`

    A string with the URL to the plugin's bug reporting tool.

    ## `repo: str`

    A string with the URL to the plugin's code repository.

    # Example

    The code below shows a `misago_plugin.py` file with a plugin manifest with all
    fields filled in:

    ```python
    from misago import MisagoPlugin


    manifest = MisagoPlugin(
        name="Example plugin with complete manifest",
        description="This plugin has all fields in its manifest filled in.",
        license="GNU GPL v2",
        icon="fa fa-wrench",
        color="#9b59b6",
        version="0.1DEV",
        author="Rafał Pitoń",
        homepage="https://misago-project.org",
        sponsor="https://github.com/sponsors/rafalp",
        help="https://misago-project.org/c/support/30/",
        bugs="https://misago-project.org/c/bug-reports/29/",
        repo="https://github.com/rafalp/misago",
    )
    ```
    """

    name: Optional[str] = None
    description: Optional[str] = None
    license: Optional[str] = None
    icon: Optional[str] = None
    color: Optional[str] = None
    version: Optional[str] = None
    author: Optional[str] = None
    homepage: Optional[str] = None
    sponsor: Optional[str] = None
    help: Optional[str] = None
    bugs: Optional[str] = None
    repo: Optional[str] = None



from unittest.mock import Mock

import pytest

from .. import SOCIALAUTH_CACHE
from ..enabledproviders import get_enabled_providers, get_provider_settings
from ..models import SocialAuthProvider


@pytest.fixture
def provider(db):
    return SocialAuthProvider.objects.create(
        provider="facebook", is_active=True, order=0
    )


def test_enabled_providers_are_read_from_db_if_cache_is_not_available(
    db, mocker, cache_versions, django_assert_num_queries
):
    mocker.patch("django.core.cache.cache.get", return_value=None)
    with django_assert_num_queries(1):
        get_enabled_providers(cache_versions)


def test_enabled_providers_are_loaded_from_cache_if_it_is_set(
    db, mocker, cache_versions, django_assert_num_queries
):
    cache_get = mocker.patch("django.core.cache.cache.get", return_value={})
    with django_assert_num_queries(0):
        get_enabled_providers(cache_versions)
    cache_get.assert_called_once()


def test_enabled_providers_cache_is_set_if_none_exists(db, mocker, cache_versions):
    cache_set = mocker.patch("django.core.cache.cache.set")
    mocker.patch("django.core.cache.cache.get", return_value=None)

    get_enabled_providers(cache_versions)
    cache_set.assert_called_once()


def test_enabled_providers_cache_is_not_set_if_it_already_exists(
    db, mocker, cache_versions, django_assert_num_queries
):
    cache_set = mocker.patch("django.core.cache.cache.set")
    mocker.patch("django.core.cache.cache.get", return_value={})
    with django_assert_num_queries(0):
        get_enabled_providers(cache_versions)
    cache_set.assert_not_called()


def test_enabled_providers_cache_key_includes_cache_name_and_version(
    db, mocker, cache_versions
):
    cache_set = mocker.patch("django.core.cache.cache.set")
    mocker.patch("django.core.cache.cache.get", return_value=None)
    get_enabled_providers(cache_versions)
    cache_key = cache_set.call_args[0][0]
    assert SOCIALAUTH_CACHE in cache_key
    assert cache_versions[SOCIALAUTH_CACHE] in cache_key


def test_enabled_providers_are_returned_as_dict(provider, mocker, cache_versions):
    mocker.patch("django.core.cache.cache.set")
    mocker.patch("django.core.cache.cache.get", return_value=None)
    providers = get_enabled_providers(cache_versions)
    assert provider.pk in providers


def test_enabled_providers_dict_includes_settings(provider, mocker, cache_versions):
    mocker.patch("django.core.cache.cache.set")
    mocker.patch("django.core.cache.cache.get", return_value=None)
    providers = get_enabled_providers(cache_versions)
    assert providers[provider.pk]["settings"]


def test_enabled_providers_dict_includes_auth_backend(provider, mocker, cache_versions):
    mocker.patch("django.core.cache.cache.set")
    mocker.patch("django.core.cache.cache.get", return_value=None)
    providers = get_enabled_providers(cache_versions)
    assert providers[provider.pk]["auth_backend"]


def test_provider_settings_keys_are_capitalized():
    provider = Mock(pk="google-oauth2", settings={"key": "text-key"})
    settings = get_provider_settings(provider)
    assert settings == {"KEY": "text-key"}


def test_provider_settings_are_merged_from_model_and_defaults():
    provider = Mock(pk="facebook", settings={"key": "text-key"})
    settings = get_provider_settings(provider)
    assert settings == {"SCOPE": ["email"], "KEY": "text-key"}



from django.urls import reverse

from ...threads import test
from ...threads.tests.test_threads_api import ThreadsApiTestCase


class UserThreadsApiTests(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()

        self.api_link = reverse("misago:api:user-threads", kwargs={"pk": self.user.pk})

    def test_invalid_user_id(self):
        """api validates user id"""
        link = reverse("misago:api:user-threads", kwargs={"pk": "abcd"})
        response = self.client.get(link)
        self.assertEqual(response.status_code, 404)

    def test_nonexistant_user_id(self):
        """api validates that user for id exists"""
        link = reverse("misago:api:user-threads", kwargs={"pk": self.user.pk + 1})
        response = self.client.get(link)
        self.assertEqual(response.status_code, 404)

    def test_empty_response(self):
        """api has no showstopers on empty response"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["results"]), 0)

    def test_user_post(self):
        """user post doesn't show in feed because its not first post in thread"""
        test.reply_thread(self.thread, poster=self.user)

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["results"]), 0)

    def test_user_event(self):
        """events don't show in feeds at all"""
        test.reply_thread(self.thread, poster=self.user, is_event=True)

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["results"]), 0)

    def test_user_thread(self):
        """user thread shows in feed"""
        thread = test.post_thread(category=self.category, poster=self.user)

        # this post will not show in feed
        test.reply_thread(thread, poster=self.user)

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["results"]), 1)
        self.assertEqual(response.json()["results"][0]["id"], thread.first_post_id)

    def test_user_thread_anonymous(self):
        """user thread shows in feed requested by unauthenticated user"""
        thread = test.post_thread(category=self.category, poster=self.user)

        self.logout_user()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["results"]), 1)
        self.assertEqual(response.json()["results"][0]["id"], thread.first_post_id)


class UserPostsApiTests(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()

        self.api_link = reverse("misago:api:user-posts", kwargs={"pk": self.user.pk})

    def test_invalid_user_id(self):
        """api validates user id"""
        link = reverse("misago:api:user-posts", kwargs={"pk": "abcd"})
        response = self.client.get(link)
        self.assertEqual(response.status_code, 404)

    def test_nonexistant_user_id(self):
        """api validates that user for id exists"""
        link = reverse("misago:api:user-posts", kwargs={"pk": self.user.pk + 1})
        response = self.client.get(link)
        self.assertEqual(response.status_code, 404)

    def test_empty_response(self):
        """api has no showstopers on empty response"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["results"]), 0)

    def test_user_event(self):
        """events don't show in feeds at all"""
        test.reply_thread(self.thread, poster=self.user, is_event=True)

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["results"]), 0)

    def test_user_hidden_post(self):
        """hidden posts don't show in feeds at all"""
        test.reply_thread(self.thread, poster=self.user, is_hidden=True)

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["results"]), 0)

    def test_user_unapproved_post(self):
        """unapproved posts don't show in feeds at all"""
        test.reply_thread(self.thread, poster=self.user, is_unapproved=True)

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["results"]), 0)

    def test_user_posts(self):
        """user posts show in feed"""
        post = test.reply_thread(self.thread, poster=self.user)
        other_post = test.reply_thread(self.thread, poster=self.user)

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["results"]), 2)
        self.assertEqual(response.json()["results"][0]["id"], other_post.pk)
        self.assertEqual(response.json()["results"][1]["id"], post.pk)

    def test_user_thread(self):
        """user thread shows in feed"""
        thread = test.post_thread(category=self.category, poster=self.user)
        post = test.reply_thread(thread, poster=self.user)

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["results"]), 2)
        self.assertEqual(response.json()["results"][0]["id"], post.pk)
        self.assertEqual(response.json()["results"][1]["id"], thread.first_post_id)

    def test_user_post_anonymous(self):
        """user post shows in feed requested by unauthenticated user"""
        post = test.reply_thread(self.thread, poster=self.user)
        other_post = test.reply_thread(self.thread, poster=self.user)

        self.logout_user()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["results"]), 2)
        self.assertEqual(response.json()["results"][0]["id"], other_post.pk)
        self.assertEqual(response.json()["results"][1]["id"], post.pk)

    def test_user_thread_anonymous(self):
        """user thread shows in feed requested by unauthenticated user"""
        thread = test.post_thread(category=self.category, poster=self.user)
        post = test.reply_thread(thread, poster=self.user)

        self.logout_user()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["results"]), 2)
        self.assertEqual(response.json()["results"][0]["id"], post.pk)
        self.assertEqual(response.json()["results"][1]["id"], thread.first_post_id)



from django.conf import settings
from django.db import models
from django.utils import timezone


class PostRead(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    category = models.ForeignKey("misago_categories.Category", on_delete=models.CASCADE)
    thread = models.ForeignKey("misago_threads.Thread", on_delete=models.CASCADE)
    post = models.ForeignKey("misago_threads.Post", on_delete=models.CASCADE)
    last_read_on = models.DateTimeField(default=timezone.now)



from hashlib import md5

from django.conf import settings
from django.core.cache import cache
from django.utils.encoding import force_bytes

from ..markup import common_flavour
from .models import Agreement, UserAgreement


def get_required_user_agreement(user, agreements):
    if user.is_anonymous:
        return None

    for agreement_type, _ in Agreement.TYPE_CHOICES:
        agreement = agreements.get(agreement_type)
        if agreement and agreement["id"] not in user.agreements:
            try:
                return Agreement.objects.get(id=agreement["id"])
            except Agreement.DoesNotExist:
                # possible stale cache
                Agreement.invalidate_cache()


def get_parsed_agreement_text(request, agreement):
    if not agreement.text:
        return None

    cache_name = "misago_legal_%s_%s" % (agreement.pk, agreement.last_modified_on or "")
    cached_content = cache.get(cache_name)

    unparsed_content = agreement.text

    checksum_source = force_bytes("%s:%s" % (unparsed_content, settings.SECRET_KEY))
    unparsed_checksum = md5(checksum_source).hexdigest()

    if cached_content and cached_content.get("checksum") == unparsed_checksum:
        return cached_content["parsed"]

    parsed = common_flavour(request, None, unparsed_content)["parsed_text"]
    cached_content = {"checksum": unparsed_checksum, "parsed": parsed}
    cache.set(cache_name, cached_content)
    return cached_content["parsed"]


def save_user_agreement_acceptance(user, agreement, commit=False):
    user.agreements.append(agreement.id)
    UserAgreement.objects.create(agreement=agreement, user=user)

    if commit:
        user.save(update_fields=["agreements"])



# Generated by Django 2.2.1 on 2019-06-09 19:13

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Icon",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("type", models.CharField(max_length=255)),
                (
                    "image",
                    models.ImageField(
                        height_field="height", upload_to="icon", width_field="width"
                    ),
                ),
                ("size", models.PositiveIntegerField(default=0)),
                ("width", models.PositiveIntegerField(default=0)),
                ("height", models.PositiveIntegerField(default=0)),
            ],
        )
    ]



import json

from django.urls import reverse

from ...acl.test import patch_user_acl
from ...conf.test import override_dynamic_settings
from ..test import AuthenticatedUserTestCase, create_test_user


class UserUsernameTests(AuthenticatedUserTestCase):
    """tests for user change name RPC (POST to /api/users/1/username/)"""

    def setUp(self):
        super().setUp()
        self.link = "/api/users/%s/username/" % self.user.pk

    @override_dynamic_settings(username_length_min=2, username_length_max=4)
    def test_get_change_username_options(self):
        """get to API returns options"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()

        self.assertIsNotNone(response_json["changes_left"])
        self.assertEqual(response_json["length_min"], 2)
        self.assertEqual(response_json["length_max"], 4)
        self.assertIsNone(response_json["next_on"])

        for i in range(response_json["changes_left"]):
            self.user.set_username("NewName%s" % i, self.user)

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["changes_left"], 0)
        self.assertIsNotNone(response_json["next_on"])

    def test_change_username_no_changes_left(self):
        """api returns error 400 if there are no username changes left"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

        for i in range(response.json()["changes_left"]):
            self.user.set_username("NewName%s" % i, self.user)

        response = self.client.get(self.link)
        self.assertEqual(response.json()["changes_left"], 0)

        response = self.client.post(self.link, data={"username": "Pointless"})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json()["detail"], "You can't change your username now."
        )
        self.assertTrue(self.user.username != "Pointless")

    def test_change_username_no_input(self):
        """api returns error 400 if new username is empty"""
        response = self.client.post(self.link, data={})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Enter new username."})

    def test_change_username_invalid_name(self):
        """api returns error 400 if new username is wrong"""
        response = self.client.post(self.link, data={"username": "####"})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "Username can only contain Latin alphabet letters, digits, "
                    "and an underscore sign."
                ),
            },
        )

    def test_change_username(self):
        """api changes username and records change"""
        response = self.client.get(self.link)
        changes_left = response.json()["changes_left"]

        old_username = self.user.username
        new_username = "NewUsernamu"

        response = self.client.post(self.link, data={"username": new_username})

        self.assertEqual(response.status_code, 200)
        options = response.json()["options"]
        self.assertEqual(changes_left, options["changes_left"] + 1)

        self.reload_user()
        self.assertEqual(self.user.username, new_username)
        self.assertTrue(self.user.username != old_username)

        self.assertEqual(self.user.namechanges.last().new_username, new_username)


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_provider="Lorem",
)
def test_change_username_api_returns_403_if_oauth_is_enabled(user, user_client):
    response = user_client.post(
        reverse("misago:api:user-username", kwargs={"pk": user.pk}),
        {
            "username": "totallyNew",
        },
    )
    assert response.status_code == 403


class UserUsernameModerationTests(AuthenticatedUserTestCase):
    """tests for moderate username RPC (/api/users/1/moderate-username/)"""

    def setUp(self):
        super().setUp()

        self.other_user = create_test_user("Other_User", "otheruser@example.com")
        self.link = "/api/users/%s/moderate-username/" % self.other_user.pk

    @patch_user_acl({"can_rename_users": 0})
    def test_no_permission(self):
        """no permission to moderate username"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't rename users."})

        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't rename users."})

    @patch_user_acl({"can_rename_users": 1})
    @override_dynamic_settings(username_length_min=3, username_length_max=12)
    def test_moderate_username(self):
        """moderate username"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

        options = response.json()
        self.assertEqual(options["length_min"], 3)
        self.assertEqual(options["length_max"], 12)

        response = self.client.post(
            self.link, json.dumps({"username": ""}), content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Enter new username."})

        response = self.client.post(
            self.link, json.dumps({"username": "$$$"}), content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "Username can only contain Latin alphabet letters, digits, "
                    "and an underscore sign."
                ),
            },
        )

        response = self.client.post(
            self.link, json.dumps({"username": "a"}), content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "Username must be at least 3 characters long."}
        )

        response = self.client.post(
            self.link,
            json.dumps({"username": "NewName"}),
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 200)

        self.other_user.refresh_from_db()
        self.assertEqual("NewName", self.other_user.username)
        self.assertEqual("newname", self.other_user.slug)

        options = response.json()
        self.assertEqual(options["username"], self.other_user.username)
        self.assertEqual(options["slug"], self.other_user.slug)

    @patch_user_acl({"can_rename_users": 1})
    def test_moderate_own_username(self):
        """moderate own username"""
        response = self.client.get("/api/users/%s/moderate-username/" % self.user.pk)
        self.assertEqual(response.status_code, 200)



from ..htmlparser import parse_html_string, print_html_string
from ..mentions import add_mentions


def test_util_replaces_mention_with_link_to_user_profile_in_parsed_text(user):
    parsing_result = {"parsed_text": f"<p>Hello, @{user.username}!</p>", "mentions": []}
    root_node = parse_html_string(parsing_result["parsed_text"])

    add_mentions(parsing_result, root_node)

    parsing_result["parsed_text"] = print_html_string(root_node)
    assert parsing_result["parsed_text"] == (
        f'<p>Hello, <a href="{user.get_absolute_url()}" '
        f'data-quote="@{user.username}">@{user.username}</a>!</p>'
    )


def test_util_adds_mention_to_parsig_result(user):
    parsing_result = {"parsed_text": f"<p>Hello, @{user.username}!</p>", "mentions": []}
    root_node = parse_html_string(parsing_result["parsed_text"])

    add_mentions(parsing_result, root_node)

    parsing_result["parsed_text"] = print_html_string(root_node)
    assert parsing_result["mentions"] == [user.id]


def test_mentions_arent_added_for_nonexisting_user(user):
    parsing_result = {"parsed_text": f"<p>Hello, @Aerith!</p>", "mentions": []}
    root_node = parse_html_string(parsing_result["parsed_text"])

    add_mentions(parsing_result, root_node)

    parsing_result["parsed_text"] = print_html_string(root_node)
    assert parsing_result["parsed_text"] == "<p>Hello, @Aerith!</p>"


def test_util_replaces_multiple_mentions_with_link_to_user_profiles_in_parsed_text(
    user, other_user
):
    parsing_result = {
        "parsed_text": f"<p>Hello, @{user.username} and @{other_user.username}!</p>",
        "mentions": [],
    }
    root_node = parse_html_string(parsing_result["parsed_text"])

    add_mentions(parsing_result, root_node)

    parsing_result["parsed_text"] = print_html_string(root_node)
    assert (
        f'<a href="{user.get_absolute_url()}" '
        f'data-quote="@{user.username}">@{user.username}</a>'
    ) in parsing_result["parsed_text"]
    assert (
        f'<a href="{other_user.get_absolute_url()}" '
        f'data-quote="@{other_user.username}">@{other_user.username}</a>'
    ) in parsing_result["parsed_text"]


def test_util_adds_multiple_mentions_to_parsing_result(user, other_user):
    parsing_result = {
        "parsed_text": f"<p>Hello, @{user.username} and @{other_user.username}!</p>",
        "mentions": [],
    }
    root_node = parse_html_string(parsing_result["parsed_text"])

    add_mentions(parsing_result, root_node)

    parsing_result["parsed_text"] = print_html_string(root_node)
    assert sorted(parsing_result["mentions"]) == sorted([user.id, other_user.id])


def test_util_handles_repeated_mentions_of_same_user(user):
    parsing_result = {
        "parsed_text": f"<p>Hello, @{user.username} and @{user.username}!</p>",
        "mentions": [],
    }
    root_node = parse_html_string(parsing_result["parsed_text"])

    add_mentions(parsing_result, root_node)

    parsing_result["parsed_text"] = print_html_string(root_node)
    assert parsing_result["mentions"] == [user.id]


def test_util_skips_mentions_in_links(user, snapshot):
    parsing_result = {
        "parsed_text": f'<p>Hello, <a href="/">@{user.username}</a></p>',
        "mentions": [],
    }
    root_node = parse_html_string(parsing_result["parsed_text"])

    add_mentions(parsing_result, root_node)

    parsing_result["parsed_text"] = print_html_string(root_node)
    assert parsing_result["parsed_text"] == (
        f'<p>Hello, <a href="/">@{user.username}</a></p>'
    )
    assert parsing_result["mentions"] == []


def test_util_handles_text_without_mentions(db):
    parsing_result = {"parsed_text": f"<p>Hello, world!</p>", "mentions": []}
    root_node = parse_html_string(parsing_result["parsed_text"])

    add_mentions(parsing_result, root_node)

    parsing_result["parsed_text"] = print_html_string(root_node)
    assert parsing_result["parsed_text"] == ("<p>Hello, world!</p>")
    assert parsing_result["mentions"] == []



from django.core.exceptions import ValidationError
from django.utils.translation import pgettext_lazy

from .utils import slugify


class validate_sluggable:
    def __init__(self, error_short=None, error_long=None):
        self.error_short = error_short or pgettext_lazy(
            "sluggable validator", "Value has to contain alpha-numerical characters."
        )
        self.error_long = error_long or pgettext_lazy(
            "sluggable validator", "Value is too long."
        )

    def __call__(self, value):
        slug = slugify(value)

        if not slug.replace("-", ""):
            raise ValidationError(self.error_short)

        if len(slug) > 255:
            raise ValidationError(self.error_long)


def validate_image_square(image):
    if image.width != image.height:
        raise ValidationError(
            pgettext_lazy("image validator", "Uploaded image is not a square.")
        )



from io import StringIO

from django.core import mail
from django.core.management import call_command

from ...conf.test import override_dynamic_settings
from ..datadownloads import request_user_data_download
from ..management.commands import prepareuserdatadownloads
from ..models import DataDownload
from ..test import AuthenticatedUserTestCase


class PrepareUserDataDownloadsTests(AuthenticatedUserTestCase):
    @override_dynamic_settings(forum_address="http://test.com/")
    def test_process_pending_data_download(self):
        """management command processes pending data download"""
        data_download = request_user_data_download(self.user)
        self.assertEqual(data_download.status, DataDownload.STATUS_PENDING)

        out = StringIO()
        call_command(prepareuserdatadownloads.Command(), stdout=out)

        command_output = out.getvalue().splitlines()[0].strip()
        self.assertEqual(command_output, "Data downloads prepared: 1")

        updated_data_download = DataDownload.objects.get(pk=data_download.pk)
        self.assertEqual(updated_data_download.status, DataDownload.STATUS_READY)
        self.assertTrue(updated_data_download.expires_on > data_download.expires_on)
        self.assertTrue(updated_data_download.file)

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(
            mail.outbox[0].subject, "TestUser, your data download is ready"
        )

        absolute_url = "".join(["http://test.com", updated_data_download.file.url])
        self.assertIn(absolute_url, mail.outbox[0].body)

    def test_skip_ready_data_download(self):
        """management command skips ready data download"""
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_READY
        data_download.save()

        out = StringIO()
        call_command(prepareuserdatadownloads.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Data downloads prepared: 0")

        updated_data_download = DataDownload.objects.get(pk=data_download.pk)
        self.assertEqual(updated_data_download.status, DataDownload.STATUS_READY)

        self.assertEqual(len(mail.outbox), 0)

    def test_skip_processing_data_download(self):
        """management command skips processing data download"""
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_PROCESSING
        data_download.save()

        out = StringIO()
        call_command(prepareuserdatadownloads.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Data downloads prepared: 0")

        updated_data_download = DataDownload.objects.get(pk=data_download.pk)
        self.assertEqual(updated_data_download.status, DataDownload.STATUS_PROCESSING)

        self.assertEqual(len(mail.outbox), 0)

    def test_skip_expired_data_download(self):
        """management command skips expired data download"""
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_EXPIRED
        data_download.save()

        out = StringIO()
        call_command(prepareuserdatadownloads.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Data downloads prepared: 0")

        updated_data_download = DataDownload.objects.get(pk=data_download.pk)
        self.assertEqual(updated_data_download.status, DataDownload.STATUS_EXPIRED)

        self.assertEqual(len(mail.outbox), 0)



from urllib.parse import urlencode

from django.contrib import messages
from django.core.paginator import EmptyPage, Paginator
from django.db import transaction
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import pgettext_lazy

from ....core.exceptions import ExplicitFirstPage
from .base import AdminView


class MassActionError(Exception):
    pass


class ListView(AdminView):
    """
    Admin items list view

    Uses following attributes:

    template = template name used to render items list
    items_per_page = number of items displayed on single page
                     (enter 0 or don't define for no pagination)
    ordering = tuple of tuples defining allowed orderings
               typles should follow this format: (name, order_by)
    """

    template_name = "list.html"

    items_per_page = 0
    ordering = None

    extra_actions = None
    mass_actions = None

    selection_label = pgettext_lazy("admin lists", "Selected: 0")
    empty_selection_label = pgettext_lazy("admin lists", "Select items")

    @classmethod
    def add_mass_action(cls, action, name, confirmation=None):
        if not cls.mass_actions:
            cls.mass_actions = []

        cls.extra_actions.append(
            {"action": action, "name": name, "confirmation": confirmation}
        )

    @classmethod
    def add_item_action(cls, name, link):
        if not cls.extra_actions:
            cls.extra_actions = []

        cls.extra_actions.append({"name": name, "link": link})

    def get_queryset(self):
        return self.get_model().objects.all()

    def dispatch(
        self, request, *args, **kwargs
    ):  # pylint: disable=too-many-branches, too-many-locals
        mass_actions_list = self.mass_actions or []
        extra_actions_list = self.extra_actions or []

        refresh_querystring = False

        context = {
            "items": self.get_queryset(),
            "paginator": None,
            "page": None,
            "order_by": [],
            "order": None,
            "filter_form": None,
            "active_filters": {},
            "querystring": "",
            "query_order": {},
            "query_filters": {},
            "selected_items": [],
            "selection_label": self.selection_label,
            "empty_selection_label": self.empty_selection_label,
            "mass_actions": mass_actions_list,
            "extra_actions": extra_actions_list,
            "extra_actions_len": len(extra_actions_list),
        }

        if request.method == "POST" and mass_actions_list:
            try:
                response = self.handle_mass_action(request, context)
                if response:
                    return response
                return redirect(request.path)
            except MassActionError as e:
                messages.error(request, e.args[0])

        if self.ordering:
            ordering_methods = self.get_ordering_methods(request)
            used_method = self.get_ordering_method_to_use(ordering_methods)
            self.set_ordering_in_context(context, used_method)

            if (
                ordering_methods["GET"]
                and ordering_methods["GET"] != ordering_methods["session"]
            ):
                # Store GET ordering in session for future requests
                session_key = self.ordering_session_key
                request.session[session_key] = ordering_methods["GET"]

            if context["order_by"] and not ordering_methods["GET"]:
                # Make view redirect to itself with querystring,
                # So address ball contains copy-friendly link
                refresh_querystring = True

        filter_form = self.get_filter_form(request)
        if filter_form:
            filtering_methods = self.get_filtering_methods(request, filter_form)
            active_filters = self.get_filtering_method_to_use(filtering_methods)
            if request.GET.get("clear_filters"):
                # Clear filters from querystring
                request.session.pop(self.filters_session_key, None)
                active_filters = {}
            self.apply_filtering_on_context(context, active_filters, filter_form)

            if (
                filtering_methods["GET"]
                and filtering_methods["GET"] != filtering_methods["session"]
            ):
                # Store GET filters in session for future requests
                session_key = self.filters_session_key
                request.session[session_key] = filtering_methods["GET"]
            if request.GET.get("set_filters"):
                # Force store filters in session
                session_key = self.filters_session_key
                request.session[session_key] = context["active_filters"]
                refresh_querystring = True

            if context["active_filters"] and not filtering_methods["GET"]:
                # Make view redirect to itself with querystring,
                # so address bar contains copy-friendly link
                refresh_querystring = True

        self.make_querystring(context)

        if self.items_per_page:
            try:
                self.paginate_items(context, kwargs.get("page", 0))
            except EmptyPage:
                return redirect(
                    "%s%s" % (reverse(self.root_link), context["querystring"])
                )

        if refresh_querystring and "redirected" not in request.GET:
            return redirect("%s%s" % (request.path, context["querystring"]))

        return self.render(request, context)

    def paginate_items(self, context, page):
        try:
            page = int(page)
            if page == 1:
                raise ExplicitFirstPage()
            elif page == 0:
                page = 1
        except ValueError:
            page = 1

        context["paginator"] = Paginator(
            context["items"], self.items_per_page, allow_empty_first_page=True
        )
        context["page"] = context["paginator"].page(page)
        context["items"] = context["page"].object_list

    # Filter list items
    filter_form = None

    def get_filter_form(self, request):
        return self.filter_form

    @property
    def filters_session_key(self):
        return "misago_admin_%s_filters" % self.root_link

    def get_filtering_methods(self, request, filter_form):
        methods = {
            "GET": self.get_filters_from_GET(request, filter_form),
            "session": self.get_filters_from_session(request, filter_form),
        }

        if request.GET.get("set_filters"):
            methods["session"] = {}

        return methods

    def get_filters_from_GET(self, request, filter_form):
        form = filter_form(request.GET)
        form.is_valid()
        return self.clean_filtering_data(form.cleaned_data)

    def get_filters_from_session(self, request, filter_form):
        session_filters = request.session.get(self.filters_session_key, {})
        form = filter_form(session_filters)
        form.is_valid()
        return self.clean_filtering_data(form.cleaned_data)

    def clean_filtering_data(self, data):
        for key, value in list(data.items()):
            if not value:
                del data[key]
        return data

    def get_filtering_method_to_use(self, methods):
        for method in ("GET", "session"):
            if methods.get(method):
                return methods.get(method)
        return {}

    def apply_filtering_on_context(self, context, active_filters, filter_form):
        context["active_filters"] = active_filters
        context["filter_form"] = filter_form(initial=context["active_filters"])

        if context["active_filters"]:
            context["items"] = context["filter_form"].filter_queryset(
                active_filters, context["items"]
            )

    # Order list items
    def get_ordering_methods(self, request):
        return {
            "GET": self.get_ordering_from_GET(request),
            "session": self.get_ordering_from_session(request),
            "default": self.clean_ordering(self.ordering[0][0]),
        }

    def get_ordering_from_GET(self, request):
        sort = request.GET.get("sort")

        if request.GET.get("direction") == "desc":
            new_ordering = "-%s" % sort
        elif request.GET.get("direction") == "asc":
            new_ordering = sort
        else:
            new_ordering = "?nope"
        return self.clean_ordering(new_ordering)

    @property
    def ordering_session_key(self):
        return "misago_admin_%s_order_by" % self.root_link

    def get_ordering_from_session(self, request):
        new_ordering = request.session.get(self.ordering_session_key)
        return self.clean_ordering(new_ordering)

    def get_ordering_method_to_use(self, methods):
        for method in ("GET", "session", "default"):
            if methods.get(method):
                return methods.get(method)

    def clean_ordering(self, new_ordering):
        for order_by, _ in self.ordering:  # pylint: disable=not-an-iterable
            if order_by == new_ordering:
                return order_by

    def set_ordering_in_context(self, context, method):
        for order_by, name in self.ordering:  # pylint: disable=not-an-iterable
            order_as_dict = {
                "type": "desc" if order_by[0] == "-" else "asc",
                "order_by": order_by,
                "name": name,
            }

            if order_by == method:
                context["order"] = order_as_dict
                context["items"] = context["items"].order_by(order_as_dict["order_by"])
            elif order_as_dict["name"]:
                if order_as_dict["type"] == "desc":
                    order_as_dict["order_by"] = order_as_dict["order_by"][1:]
                context["order_by"].append(order_as_dict)

    # Mass actions
    def handle_mass_action(self, request, context):
        limit = self.items_per_page or 64
        action = self.select_mass_action(request.POST.get("action"))
        items = [x for x in request.POST.getlist("selected_items")[:limit]]

        context["selected_items"] = items
        if not context["selected_items"]:
            raise MassActionError(
                pgettext_lazy("admin lists", "You have to select one or more items.")
            )

        action_queryset = context["items"].filter(pk__in=items)

        if not action_queryset.exists():
            raise MassActionError(
                pgettext_lazy("admin lists", "You have to select one or more items.")
            )

        action_callable = getattr(self, "action_%s" % action["action"])

        if action.get("is_atomic", True):
            with transaction.atomic():
                return action_callable(request, action_queryset)
        else:
            return action_callable(request, action_queryset)

    def select_mass_action(self, action):
        for definition in self.mass_actions:  # pylint: disable=not-an-iterable
            if definition["action"] == action:
                return definition
        raise MassActionError(pgettext_lazy("admin lists", "Action is not allowed."))

    # Querystring builder
    def make_querystring(self, context):
        values = {}
        filter_values = {}
        order_values = {}

        if context["active_filters"]:
            filter_values = context["active_filters"]
            values.update(filter_values)

        if context["order_by"]:
            order_values = {
                "sort": context["order"]["order_by"],
                "direction": context["order"]["type"],
            }

            if order_values["sort"][0] == "-":
                # We don't start sorting criteria with minus in querystring
                order_values["sort"] = order_values["sort"][1:]

            values.update(order_values)

        if values:
            values["redirected"] = 1
            context["querystring"] = "?%s" % urlencode(values, "utf-8")
        if order_values:
            context["query_order"] = order_values
        if filter_values:
            context["query_filters"] = filter_values



from django.contrib import messages
from django.utils.translation import pgettext, pgettext_lazy

from ....admin.views import generic
from ...models import Ban
from ..forms import BanForm, FilterBansForm


class BanAdmin(generic.AdminBaseMixin):
    root_link = "misago:admin:users:bans:index"
    model = Ban
    form_class = BanForm
    templates_dir = "misago/admin/bans"
    message_404 = pgettext_lazy("admin bans", "Requested ban does not exist.")

    def handle_form(self, form, request, target):
        super().handle_form(form, request, target)
        Ban.objects.invalidate_cache()


class BansList(BanAdmin, generic.ListView):
    items_per_page = 30
    ordering = [
        ("-id", pgettext_lazy("admin bans ordering choice", "From newest")),
        ("id", pgettext_lazy("admin bans ordering choice", "From oldest")),
        ("banned_value", pgettext_lazy("admin bans ordering choice", "A to z")),
        ("-banned_value", pgettext_lazy("admin bans ordering choice", "Z to a")),
    ]
    filter_form = FilterBansForm
    selection_label = pgettext_lazy("admin bans", "With bans: 0")
    empty_selection_label = pgettext_lazy("admin bans", "Select bans")
    mass_actions = [
        {
            "action": "delete",
            "name": pgettext_lazy("admin bans", "Remove bans"),
            "confirmation": pgettext_lazy(
                "admin bans", "Are you sure you want to remove those bans?"
            ),
        }
    ]

    def action_delete(self, request, items):
        items.delete()
        Ban.objects.invalidate_cache()
        messages.success(
            request, pgettext("admin bans", "Selected bans have been removed.")
        )


class NewBan(BanAdmin, generic.ModelFormView):
    message_submit = pgettext_lazy("admin bans", 'New ban "%(name)s" has been saved.')


class EditBan(BanAdmin, generic.ModelFormView):
    message_submit = pgettext_lazy("admin bans", 'Ban "%(name)s" has been edited.')


class DeleteBan(BanAdmin, generic.ButtonView):
    def button_action(self, request, target):
        target.delete()
        Ban.objects.invalidate_cache()
        message = pgettext("admin bans", 'Ban "%(name)s" has been removed.')
        messages.success(request, message % {"name": target.name})



import pytest
from django.contrib.auth import get_user_model

from ...models import DeletedUser
from ..tasks import delete_user_with_content

User = get_user_model()


def test_task_does_nothing_for_nonexisting_user_id(db):
    delete_user_with_content(1)


def test_task_does_nothing_for_staff_user_id(staffuser):
    delete_user_with_content(staffuser.id)
    staffuser.refresh_from_db()


def test_task_does_nothing_for_superuser_id(superuser):
    delete_user_with_content(superuser.id)
    superuser.refresh_from_db()


def test_task_deletes_user(user):
    delete_user_with_content(user.id)
    with pytest.raises(User.DoesNotExist):
        user.refresh_from_db()


def test_task_records_user_deletion_by_staff(user):
    delete_user_with_content(user.id)
    DeletedUser.objects.get(deleted_by=DeletedUser.DELETED_BY_STAFF)



from django.utils.module_loading import import_string

from ...conf import settings


class TreesMap:
    """Object that maps trees to strategies"""

    def __init__(self, types_modules):
        self.is_loaded = False
        self.types_modules = types_modules

    def load(self):
        self.types = self.load_types(self.types_modules)
        self.trees = self.load_trees(self.types)
        self.roots = self.get_roots(self.trees)
        self.is_loaded = True

    def load_types(self, types_modules):
        loaded_types = {}
        for path in types_modules:
            type_cls = import_string(path)
            loaded_types[type_cls.root_name] = type_cls()
        return loaded_types

    def load_trees(self, types):
        from ...categories.models import Category

        trees = {}
        for category in Category.objects.filter(level=0, special_role__in=types.keys()):
            trees[category.tree_id] = types[category.special_role]
        return trees

    def get_roots(self, trees):
        roots = {}
        for tree_id in trees:
            roots[trees[tree_id].root_name] = tree_id
        return roots

    def get_type_for_tree_id(self, tree_id):
        if not self.is_loaded:
            self.load()

        try:
            return self.trees[tree_id]
        except KeyError:
            raise KeyError("'%s' tree id has no type defined" % tree_id)

    def get_tree_id_for_root(self, root_name):
        if not self.is_loaded:
            self.load()

        try:
            return self.roots[root_name]
        except KeyError:
            raise KeyError('"%s" root has no tree defined' % root_name)


trees_map = TreesMap(settings.MISAGO_THREAD_TYPES)



import pytest
from django.contrib.auth import get_user_model

from ..models import Rank
from ..utils import hash_email

User = get_user_model()


def test_user_is_created(db):
    user = User.objects.create_user("User", "test@example.com")
    assert user.pk
    assert user.joined_on


def test_user_is_created_with_username_and_slug(db):
    user = User.objects.create_user("UserName", "test@example.com")
    assert user.slug == "username"


def test_user_is_created_with_normalized_email_and_email_hash(db):
    user = User.objects.create_user("User", "test@eXamPLe.com")
    assert user.email == "test@example.com"
    assert user.email_hash == hash_email(user.email)


def test_user_is_created_with_online_tracker(db):
    user = User.objects.create_user("User", "test@example.com")
    assert user.online_tracker
    assert user.online_tracker.last_click == user.last_login


def test_user_is_created_with_useable_password(db):
    password = "password"
    user = User.objects.create_user("UserUserame", "test@example.com", password)
    assert user.check_password(password)


def test_user_is_created_with_default_rank(db):
    user = User.objects.create_user("User", "test@example.com")
    assert user.rank == Rank.objects.get_default()


def test_user_is_created_with_custom_rank(db):
    rank = Rank.objects.create(name="Test rank")
    user = User.objects.create_user("User", "test@example.com", rank=rank)
    assert user.rank == rank


def test_newly_created_user_last_login_is_same_as_join_date(db):
    user = User.objects.create_user("User", "test@example.com")
    assert user.last_login == user.joined_on


def test_user_is_created_with_authenticated_role(db):
    user = User.objects.create_user("User", "test@example.com")
    assert user.roles.get(special_role="authenticated")


def test_user_is_created_with_diacritics_in_email(db):
    email = "łóć@łexąmple.com"
    user = User.objects.create_user("UserName", email)
    assert user.email == email


def test_creating_user_without_username_raises_value_error(db):
    with pytest.raises(ValueError):
        User.objects.create_user("", "test@example.com")


def test_creating_user_without_email_raises_value_error(db):
    with pytest.raises(ValueError):
        User.objects.create_user("User", "")


def test_create_superuser(db):
    user = User.objects.create_superuser("User", "test@example.com")
    assert user.is_staff
    assert user.is_superuser


def test_superuser_is_created_with_team_rank(db):
    user = User.objects.create_superuser("User", "test@example.com")
    assert "team" in str(user.rank)


def test_creating_superuser_without_staff_status_raises_value_error(db):
    with pytest.raises(ValueError):
        User.objects.create_superuser("User", "test@example.com", is_staff=False)


def test_creating_superuser_without_superuser_status_raises_value_error(db):
    with pytest.raises(ValueError):
        User.objects.create_superuser("User", "test@example.com", is_superuser=False)



from django.utils.deprecation import MiddlewareMixin

from ..categories.models import Category
from .models import Thread
from .viewmodels import filter_read_threads_queryset


class UnreadThreadsCountMiddleware(MiddlewareMixin):
    def process_request(self, request):
        if request.user.is_anonymous:
            return

        if not request.user_acl["can_use_private_threads"]:
            return

        if not request.user.sync_unread_private_threads:
            return

        participated_threads = request.user.threadparticipant_set.values("thread_id")

        category = Category.objects.private_threads()
        threads = Thread.objects.filter(category=category, id__in=participated_threads)

        new_threads = filter_read_threads_queryset(request, [category], "new", threads)
        unread_threads = filter_read_threads_queryset(
            request, [category], "unread", threads
        )

        request.user.unread_private_threads = (
            new_threads.count() + unread_threads.count()
        )
        request.user.sync_unread_private_threads = False

        request.user.save(
            update_fields=["unread_private_threads", "sync_unread_private_threads"]
        )



from django import forms
from django.core.exceptions import PermissionDenied
from django.utils import timezone
from django.utils.translation import npgettext, pgettext_lazy

from ...acl import algebra
from ...acl.decorators import return_boolean
from ...acl.models import Role
from ...admin.forms import YesNoSwitch
from ..models import Poll, Thread

__all__ = [
    "allow_start_poll",
    "can_start_poll",
    "allow_edit_poll",
    "can_edit_poll",
    "allow_delete_poll",
    "can_delete_poll",
    "allow_vote_poll",
    "can_vote_poll",
    "allow_see_poll_votes",
    "can_see_poll_votes",
]


class RolePermissionsForm(forms.Form):
    legend = pgettext_lazy("polls permission", "Polls")

    can_start_polls = forms.TypedChoiceField(
        label=pgettext_lazy("polls permission", "Can start polls"),
        coerce=int,
        initial=0,
        choices=[
            (0, pgettext_lazy("polls start permission choice", "No")),
            (1, pgettext_lazy("polls start permission choice", "Own threads")),
            (2, pgettext_lazy("polls start permission choice", "All threads")),
        ],
    )
    can_edit_polls = forms.TypedChoiceField(
        label=pgettext_lazy("polls permission", "Can edit polls"),
        coerce=int,
        initial=0,
        choices=[
            (0, pgettext_lazy("polls edit permission choice", "No")),
            (1, pgettext_lazy("polls edit permission choice", "Own polls")),
            (2, pgettext_lazy("polls edit permission choice", "All polls")),
        ],
    )
    can_delete_polls = forms.TypedChoiceField(
        label=pgettext_lazy("polls permission", "Can delete polls"),
        coerce=int,
        initial=0,
        choices=[
            (0, pgettext_lazy("polls delete permission choice", "No")),
            (1, pgettext_lazy("polls delete permission choice", "Own polls")),
            (2, pgettext_lazy("polls delete permission choice", "All polls")),
        ],
    )
    poll_edit_time = forms.IntegerField(
        label=pgettext_lazy(
            "polls permission", "Time limit for editing own polls, in minutes"
        ),
        help_text=pgettext_lazy(
            "polls permission", "Enter 0 to don't limit time for editing own polls."
        ),
        initial=0,
        min_value=0,
    )
    can_always_see_poll_voters = YesNoSwitch(
        label=pgettext_lazy("polls permission", "Can always see polls voters"),
        help_text=pgettext_lazy(
            "polls permission",
            "Allows users to see who voted in poll even if voting was not public.",
        ),
    )


def change_permissions_form(role):
    if isinstance(role, Role) and role.special_role != "anonymous":
        return RolePermissionsForm


def build_acl(acl, roles, key_name):
    acl.update(
        {
            "can_start_polls": 0,
            "can_edit_polls": 0,
            "can_delete_polls": 0,
            "poll_edit_time": 0,
            "can_always_see_poll_voters": 0,
        }
    )

    return algebra.sum_acls(
        acl,
        roles=roles,
        key=key_name,
        can_start_polls=algebra.greater,
        can_edit_polls=algebra.greater,
        can_delete_polls=algebra.greater,
        poll_edit_time=algebra.greater_or_zero,
        can_always_see_poll_voters=algebra.greater,
    )


def add_acl_to_poll(user_acl, poll):
    poll.acl.update(
        {
            "can_vote": can_vote_poll(user_acl, poll),
            "can_edit": can_edit_poll(user_acl, poll),
            "can_delete": can_delete_poll(user_acl, poll),
            "can_see_votes": can_see_poll_votes(user_acl, poll),
        }
    )


def add_acl_to_thread(user_acl, thread):
    thread.acl.update({"can_start_poll": can_start_poll(user_acl, thread)})


def register_with(registry):
    registry.acl_annotator(Poll, add_acl_to_poll)
    registry.acl_annotator(Thread, add_acl_to_thread)


def allow_start_poll(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("polls permission", "You have to sign in to start polls.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_close_threads": False}
    )

    if not user_acl.get("can_start_polls"):
        raise PermissionDenied(
            pgettext_lazy("polls permission", "You can't start polls.")
        )
    if user_acl.get("can_start_polls") < 2 and user_acl["user_id"] != target.starter_id:
        raise PermissionDenied(
            pgettext_lazy(
                "polls permission", "You can't start polls in other users threads."
            )
        )

    if not category_acl.get("can_close_threads"):
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "polls permission",
                    "This category is closed. You can't start polls in it.",
                )
            )
        if target.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "polls permission",
                    "This thread is closed. You can't start polls in it.",
                )
            )


can_start_poll = return_boolean(allow_start_poll)


def allow_edit_poll(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("polls permission", "You have to sign in to edit polls.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_close_threads": False}
    )

    if not user_acl.get("can_edit_polls"):
        raise PermissionDenied(
            pgettext_lazy("polls permission", "You can't edit polls.")
        )

    if user_acl.get("can_edit_polls") < 2:
        if user_acl["user_id"] != target.poster_id:
            raise PermissionDenied(
                pgettext_lazy(
                    "polls permission",
                    "You can't edit other users polls in this category.",
                )
            )
        if not has_time_to_edit_poll(user_acl, target):
            message = npgettext(
                "polls permission",
                "You can't edit polls that are older than %(minutes)s minute.",
                "You can't edit polls that are older than %(minutes)s minutes.",
                user_acl["poll_edit_time"],
            )
            raise PermissionDenied(message % {"minutes": user_acl["poll_edit_time"]})

        if target.is_over:
            raise PermissionDenied(
                pgettext_lazy(
                    "polls permission", "This poll is over. You can't edit it."
                )
            )

    if not category_acl.get("can_close_threads"):
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "polls permission",
                    "This category is closed. You can't edit polls in it.",
                )
            )
        if target.thread.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "polls permission",
                    "This thread is closed. You can't edit polls in it.",
                )
            )


can_edit_poll = return_boolean(allow_edit_poll)


def allow_delete_poll(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("polls permission", "You have to sign in to delete polls.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_close_threads": False}
    )

    if not user_acl.get("can_delete_polls"):
        raise PermissionDenied(
            pgettext_lazy("polls permission", "You can't delete polls.")
        )

    if user_acl.get("can_delete_polls") < 2:
        if user_acl["user_id"] != target.poster_id:
            raise PermissionDenied(
                pgettext_lazy(
                    "polls permission",
                    "You can't delete other users polls in this category.",
                )
            )
        if not has_time_to_edit_poll(user_acl, target):
            message = npgettext(
                "polls permission",
                "You can't delete polls that are older than %(minutes)s minute.",
                "You can't delete polls that are older than %(minutes)s minutes.",
                user_acl["poll_edit_time"],
            )
            raise PermissionDenied(message % {"minutes": user_acl["poll_edit_time"]})
        if target.is_over:
            raise PermissionDenied(
                pgettext_lazy(
                    "polls permission", "This poll is over. You can't delete it."
                )
            )

    if not category_acl.get("can_close_threads"):
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "polls permission",
                    "This category is closed. You can't delete polls in it.",
                )
            )
        if target.thread.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "polls permission",
                    "This thread is closed. You can't delete polls in it.",
                )
            )


can_delete_poll = return_boolean(allow_delete_poll)


def allow_vote_poll(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("polls permission", "You have to sign in to vote in polls.")
        )

    if target.has_selected_choices and not target.allow_revotes:
        raise PermissionDenied(
            pgettext_lazy("polls permission", "You have already voted in this poll.")
        )
    if target.is_over:
        raise PermissionDenied(
            pgettext_lazy(
                "polls permission", "This poll is over. You can't vote in it."
            )
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_close_threads": False}
    )

    if not category_acl.get("can_close_threads"):
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "polls permission", "This category is closed. You can't vote in it."
                )
            )
        if target.thread.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "polls permission", "This thread is closed. You can't vote in it."
                )
            )


can_vote_poll = return_boolean(allow_vote_poll)


def allow_see_poll_votes(user_acl, target):
    if not target.is_public and not user_acl["can_always_see_poll_voters"]:
        raise PermissionDenied(
            pgettext_lazy(
                "polls permission",
                "You dont have permission to see this poll's voters.",
            )
        )


can_see_poll_votes = return_boolean(allow_see_poll_votes)


def has_time_to_edit_poll(user_acl, target):
    edit_time = user_acl["poll_edit_time"]
    if edit_time:
        diff = timezone.now() - target.posted_on
        diff_minutes = int(diff.total_seconds() / 60)
        return diff_minutes < edit_time

    return True



import pytest

from ..providers import Providers


class Mock:
    def __init__(self, name):
        self.name = name


@pytest.fixture
def auth_backend():
    return Mock(name="facebook")


@pytest.fixture
def providers(auth_backend):
    obj = Providers()
    obj.add(
        provider="facebook",
        name="Facebook",
        auth_backend=auth_backend,
        settings={"scope": ["email"]},
        admin_form=True,
        admin_template="form.html",
    )
    return obj


def test_provider_can_be_added_to_providers(auth_backend):
    providers = Providers()
    providers.add(
        provider="facebook",
        name="Facebook",
        auth_backend=auth_backend,
        settings={"scope": ["email"]},
        admin_form=True,
        admin_template="form.html",
    )

    assert providers.dict() == {
        "facebook": {
            "provider": "facebook",
            "name": "Facebook",
            "auth_backend": auth_backend,
            "settings": {"scope": ["email"]},
            "admin_form": True,
            "admin_template": "form.html",
        }
    }
    assert providers.list() == [
        {
            "provider": "facebook",
            "name": "Facebook",
            "auth_backend": auth_backend,
            "settings": {"scope": ["email"]},
            "admin_form": True,
            "admin_template": "form.html",
        }
    ]


def test_adding_provider_fails_if_provider_key_is_different_from_backend_name(
    auth_backend,
):
    providers = Providers()
    with pytest.raises(ValueError):
        providers.add(
            provider="facebook-oauth",
            name="Facebook",
            auth_backend=auth_backend,
            settings={"scope": ["email"]},
            admin_form=True,
            admin_template="form.html",
        )

    assert providers.dict() == {}
    assert providers.list() == []


def test_providers_list_is_resorted_when_new_provider_is_added(providers, auth_backend):
    other_auth_backend = Mock(name="auth")
    providers.add(
        provider="auth",
        name="Auth",
        auth_backend=other_auth_backend,
        admin_form=True,
        admin_template="form.html",
    )

    assert providers.list() == [
        {
            "provider": "auth",
            "name": "Auth",
            "auth_backend": other_auth_backend,
            "settings": {},
            "admin_form": True,
            "admin_template": "form.html",
        },
        {
            "provider": "facebook",
            "name": "Facebook",
            "auth_backend": auth_backend,
            "settings": {"scope": ["email"]},
            "admin_form": True,
            "admin_template": "form.html",
        },
    ]


def test_util_returns_true_for_existing_provider(providers):
    assert providers.is_registered("facebook") is True


def test_util_returns_false_for_nonexisting_provider(providers):
    assert providers.is_registered("github") is False


def test_getter_returns_given_provider_name(providers):
    assert providers.get_name("facebook") == "Facebook"


def test_getter_returns_given_provider_auth_backend(providers, auth_backend):
    assert providers.get_auth_backend("facebook") is auth_backend


def test_getter_returns_given_provider_settings(providers):
    assert providers.get_settings("facebook") == {"scope": ["email"]}


def test_getter_returns_given_provider_admin_form_class(providers):
    assert providers.get_admin_form_class("facebook") is True


def test_getter_returns_given_provider_admin_template_name(providers):
    assert providers.get_admin_template_name("facebook") == "form.html"



from django.urls import reverse
from django.utils import translation

from ...test import assert_contains, assert_not_contains
from ..momentjs import MISAGO_ADMIN_MOMENT_LOCALES, get_admin_locales


def test_admin_loads_locales_list():
    assert get_admin_locales()


def test_admin_locales_list_include_some_valid_locales():
    assert "pl" in MISAGO_ADMIN_MOMENT_LOCALES
    assert "en-gb" in MISAGO_ADMIN_MOMENT_LOCALES


def test_moment_locale_is_not_included_for_english(admin_client):
    with translation.override("en"):
        response = admin_client.get(reverse("misago:admin:index"))
        assert_not_contains(response, "misago/admin/momentjs/")


def test_moment_locale_is_included_for_british_english(admin_client):
    with translation.override("en_GB"):
        response = admin_client.get(reverse("misago:admin:index"))
        assert_contains(response, "misago/admin/momentjs/en-gb.js")



from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from django.utils.translation import npgettext, pgettext
from rest_framework import serializers

from . import PostingEndpoint, PostingMiddleware
from ....acl import useracl
from ....categories import PRIVATE_THREADS_ROOT_NAME
from ....users.utils import slugify_username
from ...participants import add_participants, set_owner
from ...permissions import allow_message_user

User = get_user_model()


class ParticipantsMiddleware(PostingMiddleware):
    def use_this_middleware(self):
        if self.mode == PostingEndpoint.START:
            return self.tree_name == PRIVATE_THREADS_ROOT_NAME
        return False

    def get_serializer(self):
        return ParticipantsSerializer(
            data=self.request.data,
            context={
                "request": self.request,
                "user": self.user,
                "user_acl": self.user_acl,
            },
        )

    def post_save(self, serializer):
        set_owner(self.thread, self.user)
        add_participants(self.request.user, self.thread, serializer.users_cache)


class ParticipantsSerializer(serializers.Serializer):
    to = serializers.ListField(child=serializers.CharField(), required=True)

    def validate_to(self, usernames):
        clean_usernames = self.clean_usernames(usernames)
        self.users_cache = self.get_users(clean_usernames)

    def clean_usernames(self, usernames):
        clean_usernames = []
        for name in usernames:
            clean_name = slugify_username(name)

            if clean_name == self.context["user"].slug:
                raise serializers.ValidationError(
                    pgettext(
                        "posting api",
                        "You can't include yourself on the list of users to invite to new thread.",
                    )
                )

            if clean_name and clean_name not in clean_usernames:
                clean_usernames.append(clean_name)

        if not clean_usernames:
            raise serializers.ValidationError(
                pgettext("posting api", "You have to enter user names.")
            )

        max_participants = self.context["user_acl"]["max_private_thread_participants"]
        if max_participants and len(clean_usernames) > max_participants:
            # pylint: disable=line-too-long
            message = npgettext(
                "posting api",
                "You can't add more than %(users)s user to private thread (you've added %(added)s).",
                "You can't add more than %(users)s users to private thread (you've added %(added)s).",
                max_participants,
            )
            raise serializers.ValidationError(
                message % {"users": max_participants, "added": len(clean_usernames)}
            )

        return list(set(clean_usernames))

    def get_users(self, usernames):
        users = []
        for user in User.objects.filter(slug__in=usernames):
            try:
                user_acl = useracl.get_user_acl(
                    user, self.context["request"].cache_versions
                )
                allow_message_user(self.context["user_acl"], user, user_acl)
            except PermissionDenied as e:
                raise serializers.ValidationError(str(e))
            users.append(user)

        if len(usernames) != len(users):
            invalid_usernames = set(usernames) - {u.slug for u in users}
            sorted_usernames = sorted(invalid_usernames)

            message = pgettext(
                "posting api", "One or more users could not be found: %(usernames)s"
            )
            raise serializers.ValidationError(
                message % {"usernames": ", ".join(sorted_usernames)}
            )

        return users



from django.core.cache import cache

from ..cache.versions import invalidate_cache
from . import MENU_ITEMS_CACHE


def get_menus_cache(cache_versions):
    key = get_cache_key(cache_versions)
    return cache.get(key)


def set_menus_cache(cache_versions, menus):
    key = get_cache_key(cache_versions)
    cache.set(key, menus)


def get_cache_key(cache_versions):
    return "%s_%s" % (MENU_ITEMS_CACHE, cache_versions[MENU_ITEMS_CACHE])


def clear_menus_cache():
    invalidate_cache(MENU_ITEMS_CACHE)
